/*  Copyright 2013 Theo Berkau

    This file is part of YabauseUT

    YabauseUT is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    YabauseUT is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with YabauseUT; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//#include <lapetus.h>
#include "tests.h"
#include "dsp.h"
#include "scu.h"
#include "dsp.h"

#define SCUREG_D0R   (*(volatile u32 *)0x25FE0000)
#define SCUREG_D0W   (*(volatile u32 *)0x25FE0004)
#define SCUREG_D0C   (*(volatile u32 *)0x25FE0008)
#define SCUREG_D0AD  (*(volatile u32 *)0x25FE000C)
#define SCUREG_D0EN  (*(volatile u32 *)0x25FE0010)
#define SCUREG_D0MD  (*(volatile u32 *)0x25FE0014)

#define SCUREG_D1R   (*(volatile u32 *)0x25FE0020)
#define SCUREG_D1W   (*(volatile u32 *)0x25FE0024)
#define SCUREG_D1C   (*(volatile u32 *)0x25FE0028)
#define SCUREG_D1AD  (*(volatile u32 *)0x25FE002C)
#define SCUREG_D1EN  (*(volatile u32 *)0x25FE0030)
#define SCUREG_D1MD  (*(volatile u32 *)0x25FE0034)

#define SCUREG_D2R   (*(volatile u32 *)0x25FE0040)
#define SCUREG_D2W   (*(volatile u32 *)0x25FE0044)
#define SCUREG_D2C   (*(volatile u32 *)0x25FE0048)
#define SCUREG_D2AD  (*(volatile u32 *)0x25FE004C)
#define SCUREG_D2EN  (*(volatile u32 *)0x25FE0050)
#define SCUREG_D2MD  (*(volatile u32 *)0x25FE0054)

#define SCUREG_DSTP  (*(volatile u32 *)0x25FE0060)
#define SCUREG_DSTA  (*(volatile u32 *)0x25FE0070)

#define SCUREG_T0C   (*(volatile u32 *)0x25FE0090)
#define SCUREG_T1S   (*(volatile u32 *)0x25FE0094)
#define SCUREG_T1MD  (*(volatile u32 *)0x25FE0098)

#define SCUREG_IMS   (*(volatile u32 *)0x25FE00A0)
#define SCUREG_IST   (*(volatile u32 *)0x25FE00A4)

#define SCUREG_AIACK (*(volatile u32 *)0x25FE00A8)
#define SCUREG_ASR0  (*(volatile u32 *)0x25FE00B0)
#define SCUREG_ASR1  (*(volatile u32 *)0x25FE00B4)

#define SCUREG_RSEL  (*(volatile u32 *)0x25FE00C4)
#define SCUREG_VER   (*(volatile u32 *)0x25FE00C8)

void dsp_copy_str_to_offset(char*result);
void scu_dsp_dma_tests();

//////////////////////////////////////////////////////////////////////////////

void scu_test()
{
   int choice;

   menu_item_struct scumenu[] = {
   { "Register Test", &scu_register_test, },
   { "Interrupt Test", &scu_int_test, },
   { "DMA Test" , &scu_dma_test, },
   { "DSP Test" , &scu_dsp_test, },
   { "DSP DMA Test", &scu_dsp_dma_tests, },
//   { "Misc" , &scumisctest, },
   { "\0", NULL }
   };

   for (;;)
   {
      choice = gui_do_menu(scumenu, &test_disp_font, 0, 0, "SCU Menu", MTYPE_CENTER, -1);
      if (choice == -1)
         break;
   }   
}

//////////////////////////////////////////////////////////////////////////////

void scu_register_test(void)
{
   interrupt_set_level_mask(0xF);

   init_iapetus(RES_320x224);
   vdp_rbg0_init(&test_disp_settings);
   vdp_set_default_palette();

   // Display On
   vdp_disp_on();

   unregister_all_tests();
   register_test(&test_scu_ver_register, "VER Register");

   do_tests("SCU Register tests", 0, 0);
}

//////////////////////////////////////////////////////////////////////////////

void scu_int_test()
{
   interrupt_set_level_mask(0xF);

   init_iapetus(RES_320x224);
   vdp_rbg0_init(&test_disp_settings);
   vdp_set_default_palette();

   // Display On
   vdp_disp_on();

   unregister_all_tests();
//   register_test(&TestISTandIMS, "Interrupt status and mask");
   register_test(&test_vblank_in_interrupt, "Vblank-out Interrupt");
   register_test(&test_vblank_out_interrupt, "Vblank-in Interrupt");
   register_test(&test_hblank_in_interrupt, "HBlank-in Interrupt");
   register_test(&test_timer0_interrupt, "Timer 0 Interrupt");
   register_test(&test_timer1_interrupt, "Timer 1 Interrupt");
   register_test(&test_dsp_end_interrupt, "DSP End Interrupt");
   register_test(&test_sound_request_interrupt, "Sound Request Interrupt");
   register_test(&test_smpc_interrupt, "SMPC Interrupt");
   register_test(&test_pad_interrupt, "Pad Interrupt");
   register_test(&test_dma0_interrupt, "DMA0 Interrupt");
   register_test(&test_dma1_interrupt, "DMA1 Interrupt");
   register_test(&test_dma2_interrupt, "DMA2 Interrupt");
   register_test(&test_dma_illegal_interrupt, "Illegal DMA Interrupt");
   register_test(&test_sprite_draw_end_interrupt, "Sprite Draw End Interrupt");
//   register_test(&TestCDBlockInterrupt, "CD Block Interrupt");
   // If Netlink is connected, do a test on it too
//   if (NetlinkInit() == IAPETUS_ERR_OK)
//      register_test(&TestNetlinkInterrupt, "Netlink Interrupt");
   do_tests("SCU Interrupt tests", 0, 0);
}

//////////////////////////////////////////////////////////////////////////////

void scu_dma_test()
{
   interrupt_set_level_mask(0xF);

   init_iapetus(RES_320x224);
   vdp_rbg0_init(&test_disp_settings);
   vdp_set_default_palette();

   // Display On
   vdp_disp_on();

   unregister_all_tests();
   register_test(&test_dma0, "DMA 0 transfer");
   register_test(&test_dma1, "DMA 1 transfer");
   register_test(&test_dma2, "DMA 2 transfer");
   register_test(&test_dma_misalignment, "Misaligned DMA transfer");
   register_test(&test_indirect_dma, "Indirect DMA transfer");
//   register_test(&TestTransNum0, "0 Transfer Number transfer");
//   register_test(&TestDMAStatus, "DMA Status Register");
   // Test out DMA start factors here

//   register_test(&, );
   do_tests("SCU DMA tests", 0, 0);
}

//////////////////////////////////////////////////////////////////////////////

void test_scu_ver_register()
{
   // Make sure it's at least version 4
   if (SCUREG_VER < 4)
      stage_status = STAGESTAT_BADDATA;
   else
      stage_status = STAGESTAT_DONE;
}

//////////////////////////////////////////////////////////////////////////////

int is_int_time=0;
void (*old_func)();
u8 int_vector;
u32 old_mask;

void test_int_func()
{
   interrupt_set_level_mask(0xF);

   // Return interrupt settings back to normal
   bios_set_scu_interrupt(int_vector, old_func);
   bios_set_scu_interrupt_mask(old_mask);

   if (is_int_time)
      stage_status = STAGESTAT_DONE;
   else
      stage_status = STAGESTAT_BADINTERRUPT;

   is_int_time = 0;

   if (int_vector >= 0x50)
   {
      // Need to do an A-bus interrupt acknowledge
   }
}

//////////////////////////////////////////////////////////////////////////////

void setup_interrupt(u8 vector, u32 mask)
{
   interrupt_set_level_mask(0xF);
   int_vector = vector;
   stage_status = STAGESTAT_WAITINGFORINT;
   old_func = bios_get_scu_interrupt(int_vector);
   old_mask = bios_get_scu_interrupt_mask();
   bios_set_scu_interrupt_mask(0xFFFFFFFF);
   bios_set_scu_interrupt(int_vector, test_int_func);
   is_int_time = 1;
   bios_change_scu_interrupt_mask(~mask, 0);
}

//////////////////////////////////////////////////////////////////////////////

void test_vblank_in_interrupt()
{
   setup_interrupt(0x40, MASK_VBLANKIN);

   interrupt_set_level_mask(0xE);
}

//////////////////////////////////////////////////////////////////////////////

void test_vblank_out_interrupt()
{
   setup_interrupt(0x41, MASK_VBLANKOUT);

   interrupt_set_level_mask(0xD);
}

//////////////////////////////////////////////////////////////////////////////

void test_hblank_in_interrupt()
{
   setup_interrupt(0x42, MASK_HBLANKIN);

   interrupt_set_level_mask(0xC);
}

//////////////////////////////////////////////////////////////////////////////

void test_timer0_interrupt()
{
   setup_interrupt(0x43, MASK_TIMER0);

   // Enable Timer 0 compare
   SCUREG_T1MD = 0x001;
   SCUREG_T0C = 19;

   interrupt_set_level_mask(0xB);
}

//////////////////////////////////////////////////////////////////////////////

void test_timer1_interrupt()
{
   SCUREG_T1MD = 0;
   SCUREG_T1S = 100;

   setup_interrupt(0x44, MASK_TIMER1);
   SCUREG_T1MD = 1;

   interrupt_set_level_mask(0xA);
}

//////////////////////////////////////////////////////////////////////////////

void test_dsp_end_interrupt()
{
   u32 dsp_prog[1];

   setup_interrupt(0x45, MASK_DSPEND);

   dsp_prog[0] = ENDI();

   if (dsp_load(dsp_prog, 0, 1) != IAPETUS_ERR_OK)
   {
      stage_status = STAGESTAT_BADDATA;
      return;
   }

   dsp_exec(0);

   interrupt_set_level_mask(0x9);
}

//////////////////////////////////////////////////////////////////////////////

void test_sound_request_interrupt()
{
   setup_interrupt(0x46,  MASK_SNDREQUEST);

   // Reset all Sound interrupts
   *((volatile u16*)0x25B0042E) = 0x7FF;

   // Set MCIEB to reset when we write to bit 5 of MCIPD
   *((volatile u16*)0x25B0042A) = 0x20;

   // trigger interrupt using MCIPD
   *((volatile u16*)0x25B0042C) = 0x20;

   interrupt_set_level_mask(0x8);
}

//////////////////////////////////////////////////////////////////////////////

void test_smpc_interrupt()
{
   setup_interrupt(0x47, MASK_SYSTEMMANAGER);
   smpc_wait_till_ready();
   SMPC_REG_IREG(0) = 0x00; // no intback status
   SMPC_REG_IREG(1) = 0x0A; // 15-byte mode, peripheral data returned, time optimized
   SMPC_REG_IREG(2) = 0xF0; // ???
   smpc_issue_command(SMPC_CMD_INTBACK);
   interrupt_set_level_mask(0x7);
}

//////////////////////////////////////////////////////////////////////////////

void test_pad_interrupt()
{
   setup_interrupt(0x48, MASK_PAD);
   // Insert triggering mechanism here(fix me) - How does this one even work?
   SMPC_REG_DDR1 = 0x60;
   SMPC_REG_IOSEL = 0x1;
   SMPC_REG_EXLE = 1;
   interrupt_set_level_mask(0x7);
}

//////////////////////////////////////////////////////////////////////////////

void test_dma0_interrupt()
{
   setup_interrupt(0x4B, MASK_DMA0);

   // Do a quick DMA
   SCUREG_D0EN = 0;
   SCUREG_D0R = 0x060F0000;
   SCUREG_D0W = 0x25C40000;
   SCUREG_D0C = 0x4;
   SCUREG_D0AD = 0x101;
   SCUREG_D0MD = 0x00000007;
   SCUREG_D0EN = 0x101;

   interrupt_set_level_mask(0x4);
}

//////////////////////////////////////////////////////////////////////////////

void test_dma1_interrupt()
{
   setup_interrupt(0x4A, MASK_DMA1);

   // Do a quick DMA
   SCUREG_D1EN = 0;
   SCUREG_D1R = 0x060F0000;
   SCUREG_D1W = 0x25C40000;
   SCUREG_D1C = 0x4;
   SCUREG_D1AD = 0x101;
   SCUREG_D1MD = 0x00000007;
   SCUREG_D1EN = 0x101;

   interrupt_set_level_mask(0x5);
}

//////////////////////////////////////////////////////////////////////////////

void test_dma2_interrupt()
{
   setup_interrupt(0x49, MASK_DMA2);

   // Do a quick DMA
   SCUREG_D2EN = 0;
   SCUREG_D2R = 0x060F0000;
   SCUREG_D2W = 0x25C40000;
   SCUREG_D2C = 0x4;
   SCUREG_D2AD = 0x101;
   SCUREG_D2MD = 0x00000007;
   SCUREG_D2EN = 0x101;

   interrupt_set_level_mask(0x5);
}

//////////////////////////////////////////////////////////////////////////////

void test_dma_illegal_interrupt()
{
   setup_interrupt(0x4C, MASK_DMAILLEGAL);

   // Insert triggering mechanism here(fix me)
   // Do a quick DMA
   SCUREG_D0EN = 0;
   SCUREG_D0R = 0x060F0000;
   SCUREG_D0W = 0;
   SCUREG_D0C = 0x4;
   SCUREG_D0AD = 0x101;
   SCUREG_D0MD = 0x00000007;
   SCUREG_D0EN = 0x101;
   interrupt_set_level_mask(0x2);
}

//////////////////////////////////////////////////////////////////////////////

void test_sprite_draw_end_interrupt()
{
   sprite_struct localcoord;

   setup_interrupt(0x4D, MASK_DRAWEND);

   localcoord.attr = 0;
   localcoord.x = 0;
   localcoord.y = 0;

   vdp_start_draw_list();
   vdp_local_coordinate(&localcoord);
   vdp_end_draw_list();

   interrupt_set_level_mask(0x1);
}

//////////////////////////////////////////////////////////////////////////////

void test_cd_block_interrupt()
{
   setup_interrupt(0x50, 0x10000);

   // Unmask Subcode Q irq
   CDB_REG_HIRQ = ~HIRQ_SCDQ;
   CDB_REG_HIRQMASK = ~HIRQ_SCDQ;

   interrupt_set_level_mask(0x6);
}

//////////////////////////////////////////////////////////////////////////////

u32 dma_val_test;

void dma_int (void)
{
   bios_set_scu_interrupt(0x49, 0);
   bios_set_scu_interrupt(0x4A, 0);
   bios_set_scu_interrupt(0x4B, 0);
   bios_change_scu_interrupt_mask(0xFFFFFFFF, 0x800 | 0x400 | 0x200);

   if (*((volatile u32 *)0x25C40000) != dma_val_test)
   {
      stage_status = STAGESTAT_BADDATA;
      return;
   }

   interrupt_set_level_mask(0xF);

   stage_status = STAGESTAT_DONE;
}

//////////////////////////////////////////////////////////////////////////////

void test_dma0()
{
   dma_val_test = 0x02030405;
   *((volatile u32 *)0x260F0000) = dma_val_test;

   bios_set_scu_interrupt(0x4B, dma_int);
   bios_change_scu_interrupt_mask(~MASK_DMA0, 0);
   stage_status = STAGESTAT_WAITINGFORINT;

   // Do a quick DMA
   SCUREG_D0EN = 0;
   SCUREG_D0R = 0x060F0000;
   SCUREG_D0W = 0x25C40000;
   SCUREG_D0C = 0x4;
   SCUREG_D0AD = 0x101;
   SCUREG_D0MD = 0x00000007;
   SCUREG_D0EN = 0x101;

   interrupt_set_level_mask(0x4);
}

//////////////////////////////////////////////////////////////////////////////

void test_dma1()
{
   dma_val_test = 0x02030405;
   *((volatile u32 *)0x260F0000) = dma_val_test;

   bios_set_scu_interrupt(0x4A, dma_int);
   bios_change_scu_interrupt_mask(~MASK_DMA1, 0);
   stage_status = STAGESTAT_WAITINGFORINT;

   // Do a quick DMA
   SCUREG_D1EN = 0;
   SCUREG_D1R = 0x060F0000;
   SCUREG_D1W = 0x25C40000;
   SCUREG_D1C = 0x4;
   SCUREG_D1AD = 0x101;
   SCUREG_D1MD = 0x00000007;
   SCUREG_D1EN = 0x101;

   interrupt_set_level_mask(0x5);
}

//////////////////////////////////////////////////////////////////////////////

void test_dma2()
{
   dma_val_test = 0x02030405;
   *((volatile u32 *)0x260F0000) = dma_val_test;

   bios_set_scu_interrupt(0x49, dma_int);
   bios_change_scu_interrupt_mask(~MASK_DMA2, 0);
   stage_status = STAGESTAT_WAITINGFORINT;

   // Do a quick DMA
   SCUREG_D2EN = 0;
   SCUREG_D2R = 0x060F0000;
   SCUREG_D2W = 0x25C40000;
   SCUREG_D2C = 0x4;
   SCUREG_D2AD = 0x101;
   SCUREG_D2MD = 0x00000007;
   SCUREG_D2EN = 0x101;

   interrupt_set_level_mask(0x5);
}

//////////////////////////////////////////////////////////////////////////////

void test_dma_misalignment()
{
   *((volatile u32 *)0x260F0000) = 0x00010203;
   *((volatile u32 *)0x260F0004) = 0x04050600;
   dma_val_test = 0x02030405;

   bios_set_scu_interrupt(0x4B, dma_int);
   bios_change_scu_interrupt_mask(~MASK_DMA0, 0);
   stage_status = STAGESTAT_WAITINGFORINT;

   // Do a quick DMA
   SCUREG_D0EN = 0;
   SCUREG_D0R = 0x060F0001;
   SCUREG_D0W = 0x25C3FFFF;
   SCUREG_D0C = 0x5;
   SCUREG_D0AD = 0x101;
   SCUREG_D0MD = 0x00000007;
   SCUREG_D0EN = 0x101;

   interrupt_set_level_mask(0x4);
}

//////////////////////////////////////////////////////////////////////////////

void test_indirect_dma()
{
   static u32 dma_array[3 * 3];
   static u32 dma_data_array[] = {
      0x02030405,
      0x0708090A, 
      0x0C0D0E0F, 
   };

   interrupt_set_level_mask(0xF);

   bios_set_scu_interrupt(0x4B, dma_int);
   bios_change_scu_interrupt_mask(~MASK_DMA0, 0);
   stage_status = STAGESTAT_WAITINGFORINT;

   dma_array[0] = 1;
   dma_array[1] = 0x25C40000;
   dma_array[2] = (u32)dma_data_array;

   dma_array[3] = 2;
   dma_array[4] = 0x25C40001;
   dma_array[5] = (u32)dma_data_array+1;

//   dmaarray[6] = 1;
   dma_array[6] = 0x20000;
   dma_array[7] = 0x05C40003;
   dma_array[8] = 0x80000000 | ((u32)dma_data_array+3);

   // Do an Indirect DMA
   SCUREG_D0EN = 0;
   SCUREG_D0W = ((u32)dma_array);
   SCUREG_D0AD = 0x101;
   SCUREG_D0MD = 0x01000007;
   SCUREG_D0EN = 0x101;

   vdp_printf(&test_disp_font, 0 * 8, 21 * 8, 0xF, "DSTA = %08X", SCUREG_DSTA);

   while (SCUREG_D0EN == 0x101)
   {
      vdp_vsync();
      vdp_printf(&test_disp_font, 0 * 8, 20 * 8, 0xF, "D0EN = %08X", SCUREG_D0EN);
   }

   interrupt_set_level_mask(0x4);
}

//////////////////////////////////////////////////////////////////////////////

void test_ist_and_ims()
{
   // Mask All interrupts
   SCUREG_IMS = 0xBFFF;

   is_int_time = 0;

   // Clear any pending interrupts
   SCUREG_IST = 0;

   // Set DMA 0 interrupt function
//   bios_set_scu_interrupt(0x4B, testintfunc);
//   stagestatus = STAGESTAT_WAITINGFORINT;
   setup_interrupt(0x4B, MASK_DMA0);

   // Do a quick DMA
   SCUREG_D0R = 0x06000C00;
   SCUREG_D0W = 0x05C40000;
   SCUREG_D0C = 4;
   SCUREG_D0AD = 0x102;
   SCUREG_D0MD = 0x00010107;
   SCUREG_D0EN = 0x101;

   // Wait until DMA is finished
   while (SCUREG_DSTA & 0x13) {}

   // Unmask interrupt
   interrupt_set_level_mask(0x4);
   is_int_time = 1;
//   vdp_printf(&test_disp_font, 0 * 8, 10 * 8, 0xF, "%04X", SCUREG_IST);
   SCUREG_IMS = 0xB7FF;

/*
   // wait a while
   for (i = 0; i < (60 * 2); i++)
   {
      wait_vblank_out();
      wait_vblank_in();

      vdp_printf(&test_disp_font, 0 * 8, 11 * 8, 0xF, "%04X", SCUREG_IST);
   }
   interrupt_set_level_mask(0x4);
*/
}

//////////////////////////////////////////////////////////////////////////////

int dma_print_pos = 0;
volatile int dma_finished = 0;
volatile int interrupt_error = 0;
volatile int num_interrupts = 0;

struct DmaStruct
{
   u32 source_addr;
   u32 dest_addr;
   u32 length;
   u32 read_add;
   u32 write_add;
   u32 factor;
   u32 enable;
   u32 level;
};

//////////////////////////////////////////////////////////////////////////////

void frc_clear()
{
   SH2_REG_FRC_W(0);
}

//////////////////////////////////////////////////////////////////////////////

int frc_get()
{
   return SH2_REG_FRC_R;
}

//////////////////////////////////////////////////////////////////////////////

void do_dma(u32 src_addr, u32 dst_addr, u32 read_add, u32 write_add, u32 length, u32 print_result, u32 factor)
{
   dma_finished = 0;

   SCUREG_D0EN = 0;
   SCUREG_D0R = src_addr;
   SCUREG_D0W = dst_addr;
   SCUREG_D0C = length;
   SCUREG_D0AD = (read_add << 8) | write_add;
   SCUREG_D0MD = factor;

   frc_clear();

   SCUREG_D0EN = 0x101;

   while (!dma_finished) {}

   u32 endtime = frc_get();

   if (print_result)
      vdp_printf(&test_disp_font, 0 * 8, dma_print_pos * 8, 0xF, "frc: %d (~%d cycles)", endtime, endtime * 8);

   dma_print_pos++;
}

//////////////////////////////////////////////////////////////////////////////

void scu_dma_memset(u32 destination, u32 length)
{

   volatile u32 *p = (volatile u32 *)(0x260F0000);
   p[0] = 0;
   do_dma(0x260F0000, destination, 0, 1, length, 0, 7);
}

//////////////////////////////////////////////////////////////////////////////

void memset_test_areas(u32 length)
{
   scu_dma_memset(0x25E00000, 320 * 224 * 2);
   scu_dma_memset(0x25a00000, length);
   scu_dma_memset(0x25C00000, length);
}

//////////////////////////////////////////////////////////////////////////////

void dma_interrupt()
{
   if (dma_finished == 1)
      interrupt_error = 1;

   dma_finished = 1;

   num_interrupts += 1;
}

//////////////////////////////////////////////////////////////////////////////

void dma_interrupt_setup()
{
   bios_set_scu_interrupt(0x4B, dma_interrupt);
   bios_change_scu_interrupt_mask(~MASK_DMA0, 0);
   interrupt_set_level_mask(0x4);
}

//////////////////////////////////////////////////////////////////////////////

#define RSTCSR_W (*(volatile u16 *)0XFFFFFE82)

#define WRITE_DEST(DESTINATION_ADDR) \
   *DESTINATION_ADDR = 1; \
   *DESTINATION_ADDR = 2; \
   *DESTINATION_ADDR = 3; \
   *DESTINATION_ADDR = 2;

//////////////////////////////////////////////////////////////////////////////

void write_timing(u32 destination, char*test_name)
{
   memset_test_areas(0x1000);

#ifdef WANT_32
   volatile u32 *dest_ptr = (volatile u32 *)(destination);
#else
   volatile u16 *dest_ptr = (volatile u16 *)(destination);
#endif

   dma_print_pos = 10;

   int i;

   int num_writes = 16;

   vdp_printf(&test_disp_font, 0 * 8, 1 * 8, 0xF, "%s", test_name);

   for (i = 0; i < 5; i++)
   {
      vdp_wait_vblankin();

      //zero watchdog timer
      SH2_REG_WTCNT_W(0);
      RSTCSR_W = 0;

      //enable timer
      SH2_REG_WTCSR_W(1 << 5);

      WRITE_DEST(dest_ptr);
      WRITE_DEST(dest_ptr);
      WRITE_DEST(dest_ptr);
      WRITE_DEST(dest_ptr);

      u8 end_time = SH2_REG_WTCNT_R;

      //disable timer
      SH2_REG_WTCSR_W(0);

      frc_clear();

      WRITE_DEST(dest_ptr);
      WRITE_DEST(dest_ptr);
      WRITE_DEST(dest_ptr);
      WRITE_DEST(dest_ptr);

      u32 endtime = frc_get();

      vdp_printf(&test_disp_font, 0 * 8, dma_print_pos * 8, 0xF, "frc: %d (~%d cycles), ~%d cycles per write", endtime, endtime * 8, (endtime * 8) / num_writes);
      vdp_printf(&test_disp_font, 0 * 8, (dma_print_pos + 6) * 8, 0xF, "wdt: %d (~%d cycles), ~%d cycles per write", end_time, end_time * 2, (end_time * 2) / num_writes);

      dma_print_pos++;
   }

   for (;;)
   {
      vdp_wait_vblankin();

      if (per[0].but_push_once & PAD_A)
      {
         break;
      }

      if (per[0].but_push_once & PAD_Y)
      {
         reset_system();
      }
   }
}

//////////////////////////////////////////////////////////////////////////////

void write_timing_test()
{
   dma_interrupt_setup();

   write_timing(0x25E00000, "to vdp2 vram");

   write_timing(0x25a00000, "to scsp ram");

   write_timing(0x25C00000, "to vdp1 ram");

   write_timing(0x25C80000, "to vdp1 framebuffer");

   write_timing(0x20200000, "to low work ram");

   write_timing(0x260FF000, "to high work ram");
}

//////////////////////////////////////////////////////////////////////////////

void print_data(u32 x_start, u32 y_start, u32 addr, u32 length, char * str)
{
   volatile u32 *ptr = (volatile u32 *)(addr);
   int i;

   vdp_printf(&test_disp_font, x_start * 8, y_start * 8, 0xF, str);

   vdp_printf(&test_disp_font, x_start * 8, (y_start + 1) * 8, 0xF, "start");

   for (i = 0; i < 16; i++)
   {
      vdp_printf(&test_disp_font, x_start * 8, (i + y_start + 2) * 8, 0xF, "%08X", ptr[i]);
   }

   vdp_printf(&test_disp_font, (x_start + 9) * 8, (y_start + 1) * 8, 0xF, "end");

   for (i = 0; i < 16; i++)
   {
      vdp_printf(&test_disp_font, (x_start + 9) * 8, (i + y_start + 2) * 8, 0xF, "%08X", ptr[i + ((length / 4) - 16)]);
   }
}

//////////////////////////////////////////////////////////////////////////////

void memory_setup(u32 addr1, u32 addr2, u32 length)
{
   int i;

   volatile u32 *dest_1 = (volatile u32*)(addr1);
   volatile u32 *dest_2 = (volatile u32*)(addr2);

   for (i = 0; i < (length / 4); i += 4)
   {
      dest_1[i + 0] = 0x00000008;
      dest_1[i + 1] = 0x00010009;
      dest_1[i + 2] = 0x0002000a;
      dest_1[i + 3] = 0x0003000b;
   }

   for (i = 0; i < (length / 4); i += 4)
   {
      dest_2[i + 0] = 0x0004000c;
      dest_2[i + 1] = 0x0005000d;
      dest_2[i + 2] = 0x0006000e;
      dest_2[i + 3] = 0x0007000f;
   }
}

//////////////////////////////////////////////////////////////////////////////

void finish_loop()
{
   for (;;)
   {
      vdp_vsync();

      if (interrupt_error)
      {
         vdp_printf(&test_disp_font, 0 * 8, 0 * 8, 0xF, "error");
      }

      if (per[0].but_push_once & PAD_A)
      {
         break;
      }

      if (per[0].but_push_once & PAD_Y)
      {
         reset_system();
      }
   }
}

//////////////////////////////////////////////////////////////////////////////

void test_boilerplate(u32 src_addr, u32 length, char* test_name)
{
   memset_test_areas(length);

   memory_setup(src_addr, src_addr + length, length);

   vdp_printf(&test_disp_font, 0 * 8, 2 * 8, 0xF, "%s", test_name);

   //source data
   print_data(0, 3, src_addr, length, "source");

   dma_print_pos = 22;
}

//////////////////////////////////////////////////////////////////////////////

void test_direct_dma(u32 src_addr, u32 dst_addr, u32 read_add, u32 write_add, u32 factor, u32 length, char* test_name)
{
   test_boilerplate(src_addr, length, test_name);

   int i;
   for (i = 0; i < 5; i++)
   {
      //vdp2 dma takes much longer during active display
      vdp_wait_vblankin();
      do_dma(src_addr, dst_addr, read_add, write_add, length, 1, factor);
   }

   //destination data
   print_data(18, 3, dst_addr, length, "destination");

   finish_loop();
}

//////////////////////////////////////////////////////////////////////////////

void do_dma_multiple_struct(
struct DmaStruct * dma_struct,
   u32 print_result,
   u32 num_dmas)
{
   dma_finished = 0;

   int i;

   SCUREG_D0EN = 0;

   for (i = 0; i < num_dmas; i++)
   {
      u32 level = dma_struct[i].level;

      if (level == 0)
      {
         SCUREG_D0R = dma_struct[i].source_addr;
         SCUREG_D0W = dma_struct[i].dest_addr;
         SCUREG_D0C = dma_struct[i].length;
         SCUREG_D0AD = (dma_struct[i].read_add << 8) | dma_struct[i].write_add;
         SCUREG_D0MD = dma_struct[i].factor;
         SCUREG_D0EN = dma_struct[i].enable;
      }
      else if (level == 1)
      {
         SCUREG_D1R = dma_struct[i].source_addr;
         SCUREG_D1W = dma_struct[i].dest_addr;
         SCUREG_D1C = dma_struct[i].length;
         SCUREG_D1AD = (dma_struct[i].read_add << 8) | dma_struct[i].write_add;
         SCUREG_D1MD = dma_struct[i].factor;
         SCUREG_D1EN = dma_struct[i].enable;
      }
      else if (level == 2)
      {
         SCUREG_D2R = dma_struct[i].source_addr;
         SCUREG_D2W = dma_struct[i].dest_addr;
         SCUREG_D2C = dma_struct[i].length;
         SCUREG_D2AD = (dma_struct[i].read_add << 8) | dma_struct[i].write_add;
         SCUREG_D2MD = dma_struct[i].factor;
         SCUREG_D2EN = dma_struct[i].enable;
      }
   }

   frc_clear();

   while (!dma_finished) {}

   u32 endtime = frc_get();

   if (print_result)
      vdp_printf(&test_disp_font, 0 * 8, dma_print_pos * 8, 0xF, "frc: %d (~%d cycles)", endtime, endtime * 8);

   dma_print_pos++;
}

//////////////////////////////////////////////////////////////////////////////

void test_multiple_direct_dma(u32 dst_addr, u32 read_add, u32 write_add, char* test_name, u32 num_dmas)
{
   u32 length = 0x800;
   memset_test_areas(length);

   vdp_printf(&test_disp_font, 0 * 8, 2 * 8, 0xF, "%s", test_name);

   u32 src_addrs[32] = { 0 };
   u32 dst_addrs[32] = { 0 };
   u32 data[32] = { 0 };

   struct DmaStruct dma_structs[16] = { { 0 } };

   int i;

   for (i = 0; i < num_dmas; i++)
   {
      dma_structs[i].source_addr = 0x260F0000 + length*i;
      dma_structs[i].dest_addr = 0x25E00000 + length*i;
      dma_structs[i].length = length;
      dma_structs[i].read_add = read_add;
      dma_structs[i].write_add = write_add;
      dma_structs[i].factor = 7;
      dma_structs[i].enable = 0x101;
      dma_structs[i].level = 1;

      src_addrs[i] = 0x260F0000 + length*i;
      dst_addrs[i] = 0x25E00000 + length*i;
      data[i] = 0xcafe0000 + i;
   }

   //source data
   print_data(0, 3, src_addrs[0], length, "source");

   dma_print_pos = 22;

   int j;

   for (i = 0; i < num_dmas; i++)
   {
      volatile u32 *dst = (volatile u32 *)(src_addrs[i]);

      for (j = 0; j < (length / 4); j += 4)
      {
         dst[j + 0] = data[i];
         dst[j + 1] = data[i];
         dst[j + 2] = data[i];
         dst[j + 3] = data[i];
      }
   }

   num_interrupts = 0;

   vdp_wait_vblankin();

   do_dma_multiple_struct(dma_structs, 1, num_dmas);

   for (i = 0; i < num_dmas; i++)
   {
      volatile u32 *ptr = (volatile u32 *)(dst_addrs[i]);
      u32 x_start = 18;
      u32 y_start = 3;
      vdp_printf(&test_disp_font, (x_start + 9) * 8, (i + y_start + 2) * 8, 0xF, "%08X", ptr[(length / 4) - 1]);
   }

   vdp_printf(&test_disp_font, 18 * 8, 27 * 8, 0xF, "%08X", num_interrupts);

   finish_loop();
}

//////////////////////////////////////////////////////////////////////////////

void test_all_direct_dma()
{
   u32 high_wram_addr = 0x260F0000;
   u32 vdp2_vram_addr = 0x25E00000;
   u32 scsp_ram_addr = 0x25a00000;
   u32 vdp1_ram_addr = 0x25C00000;
   u32 vdp1_fb_addr = 0x25C80000;

   u32 length = 0x1000;

   int i;

   test_multiple_direct_dma(vdp2_vram_addr, 1, 1, "multiple dma %d", 1);

   for (i = 1; i < 17; i++)
   {
      test_multiple_direct_dma(vdp2_vram_addr, 1, 1, "multiple dma %d", i);
   }

   test_direct_dma(high_wram_addr, vdp2_vram_addr, 1, 1, 7, length, "wram to vdp2 vram");
   test_direct_dma(high_wram_addr, scsp_ram_addr, 1, 1, 7, length, "wram to scsp ram");
   test_direct_dma(high_wram_addr, vdp1_ram_addr, 1, 1, 7, length, "wram to vdp1 ram");

   vdp_vsync();
   VDP1_REG_FBCR = 3;
   vdp_vsync();

   test_direct_dma(high_wram_addr, vdp1_fb_addr, 1, 1, 7, length, "wram to vdp1 fb");

   test_direct_dma(vdp2_vram_addr, high_wram_addr, 1, 2, 7, length, "vdp2 vram to wram");
   test_direct_dma(scsp_ram_addr, high_wram_addr, 1, 2, 7, length, "scsp ram to wram");
   test_direct_dma(vdp1_ram_addr, high_wram_addr, 1, 2, 7, length, "vdp1 ram to wram");

   vdp_vsync();
   VDP1_REG_FBCR = 3;
   vdp_vsync();

   test_direct_dma(vdp1_fb_addr, high_wram_addr, 1, 2, 7, length, "vdp1 fb to wram");

   vdp_vsync();
   VDP1_REG_FBCR = 0;
   vdp_vsync();
}

//////////////////////////////////////////////////////////////////////////////

void indirect_dma_test()
{
   int length = 0x1000;

   int i;

   volatile u32 *dest_ptr = (volatile u32 *)(0x25E00000);
   volatile u32 *dest_ptr_len = (volatile u32 *)(0x25E02000);

   memory_setup(0x260F0000, 0x260F2000, length);

   u32 indirect_table_addr = 0x260FF000;

   volatile u32 *indirect_table = (volatile u32 *)(indirect_table_addr);

   indirect_table[0] = length;
   indirect_table[1] = 0x25E00000;
   indirect_table[2] = 0x260F0000;

   indirect_table[3] = length;
   indirect_table[4] = 0x25E02000;
   indirect_table[5] = 0x260F2000 | 0x80000000;

   bios_set_scu_interrupt(0x4B, dma_interrupt);
   bios_change_scu_interrupt_mask(~MASK_DMA0, 0);
   dma_finished = 0;

   SCUREG_D0W = indirect_table_addr;
   SCUREG_D0AD = 0x101;
   SCUREG_D0MD = (1 << 24) | 0x00000007;
   SCUREG_D0EN = 0x101;

   interrupt_set_level_mask(0x4);

   while (!dma_finished) {}

   for (i = 0; i < 16; i++)
   {
      vdp_printf(&test_disp_font, 0 * 8, (i + 4) * 8, 0xF, "%08X", dest_ptr[i]);
   }

   for (i = 0; i < 16; i++)
   {
      vdp_printf(&test_disp_font, 12 * 8, (i + 4) * 8, 0xF, "%08X", dest_ptr_len[i]);
   }

   finish_loop();
}

//////////////////////////////////////////////////////////////////////////////

void dma_update_test(u32 write_address_update, u32 read_address_update, u32 write_add, u32 read_add)
{
   int i;

   u32 src_addr = 0x260F0000;
   u32 dest_addr = 0x25E00000;

   int length = 0x1000;

   volatile u32 *dest_ptr = (volatile u32 *)(dest_addr);
   volatile u32 *dest_ptr2 = (volatile u32 *)(dest_addr + length);

   memset_test_areas(length);

   memory_setup(src_addr, src_addr + length, length);

   bios_set_scu_interrupt(0x4B, dma_interrupt);
   bios_change_scu_interrupt_mask(~MASK_DMA0, 0);

   SCUREG_D0R = src_addr;
   SCUREG_D0W = dest_addr;
   SCUREG_D0C = length;
   SCUREG_D0AD = (read_add << 8) | write_add;
   SCUREG_D0MD = 0x00000007 | (write_address_update << 8) | (read_address_update << 16);
   SCUREG_D0EN = 0x101;

   interrupt_set_level_mask(0x4);

   dma_finished = 0;

   while (!dma_finished) {}

   volatile u32 *ack = (volatile u32 *)(0x25fe00a4);

   ack[0] = 0;

   SCUREG_D0EN = 0x101;

   dma_finished = 0;

   for (i = 0; i < 16; i++)
   {
      vdp_printf(&test_disp_font, 0 * 8, (i + 4) * 8, 0xF, "%08X", dest_ptr[i]);
   }

   for (i = 0; i < 16; i++)
   {
      vdp_printf(&test_disp_font, 12 * 8, (i + 4) * 8, 0xF, "%08X", dest_ptr2[i]);
   }

   finish_loop();
}

//submit upper level dma while lower is running
void dma_level_priority_impl(u32 src_addr, u32 dest_addr, u32 write_add)
{
   u32 read_add = 0;
   u32 write_address_update = 0;
   u32 read_address_update = 0;
   int i;
   u32 inc = 0x6000;
   for (i = 0; i < 0x10000; i++)
      ((volatile u32 *)(dest_addr))[i] = 0;

   vdp_printf(&test_disp_font, 18 * 8, 16 * 8, 0xF, "%08X %08X", ((volatile u32 *)(src_addr))[0], ((volatile u32 *)(src_addr))[1]);
   vdp_printf(&test_disp_font, 18 * 8, 17 * 8, 0xF, "%04X %04X", ((volatile u16 *)(src_addr))[0], ((volatile u16 *)(src_addr))[1]);

   if (src_addr != 0x22000000)
   {
      volatile u32 *src = (volatile u32 *)(src_addr);

      src[0] = 0xdeadbeef;
      src[1] = 0xaaaaaaaa;
   }

   SCUREG_D0R = src_addr;
   SCUREG_D0W = dest_addr;
   SCUREG_D0C = 0x800;
   SCUREG_D0AD = (read_add << 8) | write_add;
   SCUREG_D0MD = 0x00000007 | (write_address_update << 8) | (read_address_update << 16);
   SCUREG_D0EN = 0x101;

   SCUREG_D1R = src_addr + 4;
   SCUREG_D1W = dest_addr + inc;
   SCUREG_D1C = 0x800;
   SCUREG_D1AD = (read_add << 8) | write_add;
   SCUREG_D1MD = 0x00000007 | (write_address_update << 8) | (read_address_update << 16);
   SCUREG_D1EN = 0x101;

   vdp_vsync();
   vdp_vsync();
   vdp_vsync();

   for (i = 0; i < 16; i++)
   {
      vdp_printf(&test_disp_font, 0 * 8, (i + 16) * 8, 0xF, "%08X", ((volatile u32 *)(dest_addr))[i]);
      vdp_printf(&test_disp_font, 9 * 8, (i + 16) * 8, 0xF, "%08X", ((volatile u32 *)(dest_addr + inc))[i]);
   }

   for (;;)
   {
      vdp_vsync();

      if (per[0].but_push_once & PAD_A)
      {
         break;
      }

      if (per[0].but_push_once & PAD_Y)
      {
         reset_system();
      }
   }
}

void dma_level_priority_indirect(u32 src_addr, u32 dest_addr, u32 write_add)
{
   u32 read_add = 0;
   u32 write_address_update = 0;
   u32 read_address_update = 0;
   int i;
   u32 inc = 0x6000;
   for (i = 0; i < 0x10000; i++)
      ((volatile u32 *)(dest_addr))[i] = 0;

   if (src_addr != 0x22000000)
   {
      volatile u32 *src = (volatile u32 *)(src_addr);

      src[0] = 0xdeadbeef;
      src[1] = 0xaaaaaaaa;
   }

   u32 indirect_table_addr_0 = 0x260FE000;

   volatile u32 *indirect_table_0 = (volatile u32 *)(indirect_table_addr_0);

   u32 length = 0x400;

   indirect_table_0[0] = length;//size
   indirect_table_0[1] = dest_addr;//write addr
   indirect_table_0[2] = src_addr;//read addr

   indirect_table_0[3] = length;
   indirect_table_0[4] = dest_addr + 0x3000;
   indirect_table_0[5] = src_addr + 4 | 0x80000000;

   u32 indirect_table_addr_1 = 0x260FF010;

   volatile u32 *indirect_table_1 = (volatile u32 *)(indirect_table_addr_1);

   indirect_table_1[0] = length;//size
   indirect_table_1[1] = dest_addr + 0x6000;//write addr
   indirect_table_1[2] = src_addr;//read addr

   indirect_table_1[3] = length;
   indirect_table_1[4] = dest_addr + 0x9000;
   indirect_table_1[5] = src_addr + 4 | 0x80000000;

   SCUREG_D0W = indirect_table_addr_0;
   SCUREG_D0AD = 0x101;
   SCUREG_D0MD = (1 << 24) | 0x00000007;
   SCUREG_D0EN = 0x101;

   SCUREG_D1W = indirect_table_addr_1;
   SCUREG_D1AD = 0x101;
   SCUREG_D1MD = (1 << 24) | 0x00000007;
   SCUREG_D1EN = 0x101;

   vdp_vsync();
   vdp_vsync();
   vdp_vsync();

   for (i = 0; i < 16; i++)
   {
      vdp_printf(&test_disp_font, 0 * 8, (i + 16) * 8, 0xF, "%08X", ((volatile u32 *)(dest_addr))[i]);
      vdp_printf(&test_disp_font, 9 * 8, (i + 16) * 8, 0xF, "%08X", ((volatile u32 *)(dest_addr + inc))[i]);
   }

   for (;;)
   {
      vdp_vsync();

      if (per[0].but_push_once & PAD_A)
      {
         break;
      }

      if (per[0].but_push_once & PAD_Y)
      {
         reset_system();
      }
   }
}

void dma_level_priority()
{
   int i;

   dma_level_priority_indirect(0x260F0000, 0x25E00000, 1);

   for (i = 0; i < 7; i++)
      dma_level_priority_impl(0x260F0000, 0x25E00000, i);

   //dma from cartridge area
   for (i = 0; i < 7; i++)
      dma_level_priority_impl(0x22000000, 0x25E00000, i);
}

#undef SCUREG_DSTA
#define SCUREG_DSTA  (*(volatile u32 *)0x25fe007c)

#define NUM_VALUES 4096

u32 values[NUM_VALUES] = { 0 };

void do_flag_dma(u32 src_addr, u32 dst_addr, u32 read_add, u32 write_add, u32 length)
{
   SCUREG_D0EN = 0;
   SCUREG_D0R = src_addr;
   SCUREG_D0W = dst_addr;
   SCUREG_D0C = length;
   SCUREG_D0AD = (read_add << 8) | write_add;
   SCUREG_D0MD = 7;

   int pos = 0;

   u32* src = (volatile u32*)(src_addr);
   u32* dest = (volatile u32*)(dst_addr);

   int i;

   for (i = 0; i < length; i+=2)
   {
      src[i] = 0xdeadbeef;
      src[i+1] = 0xcafef00d;
      dest[i+0] = 0;
      dest[i + 1] = 0;
   }

   frc_clear();

   u32 first = 0;
   int was_gotten = 0;

   SCUREG_D0EN = 0x101;

   for (;;)
   {
      u32 val = SCUREG_DSTA;

      if (val)
      {
         first = frc_get();
         was_gotten = 1;
      }
         
      values[pos] = val;

      if (pos < 4096)
         pos++;
      else
         break;
   }

   for(i = 0; i < 27; i++)
   {
      vdp_printf(&test_disp_font, 32 * 8, i * 8, 0xF, "%08X", values[i]);
   }

   for (i = 0; i < 27; i++)
   {
      vdp_printf(&test_disp_font, 16 * 8, i * 8, 0xF, "%08X", dest[i]);
   }

   vdp_printf(&test_disp_font, 0 * 8, 0 * 8, 0xF, "%d %d", pos, was_gotten);

   vdp_printf(&test_disp_font, 0 * 8, 1 * 8, 0xF, "frc: %d (~%d cycles)", first, first * 8);

   for (;;)
   {
      vdp_vsync();

      if (per[0].but_push_once & PAD_A)
      {
         break;
      }

      if (per[0].but_push_once & PAD_Y)
      {
         reset_system();
      }
   }
}

#define AUTO 1

u32 c_results[128][65] =
{
{0xdead0000, 0xdead0000, 0xcafef00d, 0xcafef00d, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000},
{ 0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
{ 0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0000,0x00000000,0xcafef00d,0xdead0002,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0000,0x00000000,0x00000000,0x00000000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0xdead0002,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0000,0x00000000,0xcafef00d,0x00000000,0xdead0002,0xcafef00d,0x00000000,0xdead0004,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0002,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0002,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0x00000000,0xdead0004,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0x00000002,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0x00000002,0x00000000,0xcafef00d,0xdead0002,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0x00000002,0x00000000,0x00000000,0x00000000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0x00000002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0x00000002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0x00000002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0xdead0002,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0x00000003,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0x00000003,0x00000000,0xcafef00d,0x00000000,0xdead0002,0xcafef00d,0x00000000,0xdead0004,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0x00000003,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0x00000003,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0x00000003,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0x00000003,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0002,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0002,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0x00000000,0xdead0004,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
{ 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
};

u32 b_results[128][65] =
{
   //b-bus
   { 0xdead0000,0xcafef00d,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xf00d0000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0xcafe0000,0xdead0000,0x00020000,0xcafe0000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0xcafe0000,0xdead0000,0xf00d0000,0x00020000,0x00000000,0xcafe0000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafe0000,0xdead0000,0x00000000,0x00000000,0xf00d0000,0x00020000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafe0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00020000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafe0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00020000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00020000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0xcafe0000,0xf00d0000,0xdead0000,0xcafe0000,0xf00d0000,0xdead0000,0x00040000,0xcafe0000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0xf00d0000,0x00000000,0xdead0000,0xcafe0000,0x00020000,0xf00d0000,0x00000000,0xdead0000,0x00000000,0x00040000,0x00000000,0xcafe0000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0xdead0000,0xcafe0000,0x00000000,0x00000000,0x00020000,0xf00d0000,0x00000000,0x00000000,0x00000000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00040000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0000,0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00020000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00040000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00020000,0xcafe0000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00020000,0x00000000,0xcafe0000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00020000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00020000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00020000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00020000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafe0000,0xf00d0000,0xdead0000,0x00040000,0xcafe0000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafe0000,0x00000000,0xf00d0000,0x00000000,0xdead0000,0x00000000,0x00040000,0x00000000,0xcafe0000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafe0000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00040000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00040000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00040000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xf00d0000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000002,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000000,0x00020000,0xcafe0000,0xdead0000,0x00020000,0xcafe0000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000000,0x00000000,0x00020000,0x00000000,0xcafe0000,0xdead0000,0xf00d0000,0x00020000,0x00000000,0xcafe0000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000000,0x00000000,0x00000000,0x00000000,0x00020000,0x00000000,0x00000000,0x00000000,0xcafe0000,0xdead0000,0x00000000,0x00000000,0xf00d0000,0x00020000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00020000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafe0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00020000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00020000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafe0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00020000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00020000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00020000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000003,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000000,0x00030000,0xcafe0000,0xf00d0000,0xdead0000,0xcafe0000,0xf00d0000,0xdead0000,0x00040000,0xcafe0000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000000,0x00000000,0x00030000,0x00000000,0xcafe0000,0x00000000,0xf00d0000,0x00000000,0xdead0000,0xcafe0000,0x00020000,0xf00d0000,0x00000000,0xdead0000,0x00000000,0x00040000,0x00000000,0xcafe0000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000000,0x00000000,0x00000000,0x00000000,0x00030000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0xdead0000,0xcafe0000,0x00000000,0x00000000,0x00020000,0xf00d0000,0x00000000,0x00000000,0x00000000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00040000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00030000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0000,0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00020000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00040000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00030000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00030000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00020000,0xcafe0000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00020000,0x00000000,0xcafe0000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00020000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00020000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00020000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00020000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafe0000,0xf00d0000,0xdead0000,0x00040000,0xcafe0000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafe0000,0x00000000,0xf00d0000,0x00000000,0xdead0000,0x00000000,0x00040000,0x00000000,0xcafe0000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafe0000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00040000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00040000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00040000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafe0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xf00d0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
};

u32 a_results[128][65] =
{
   //a-bus
   { 0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xcafef00d,0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xdead0000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000 },
   { 0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0xcafef00d,0xdead0002,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0xdead0002,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0xcafef00d,0x00000000,0xdead0002,0xcafef00d,0x00000000,0xdead0004,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0002,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0002,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0x00000000,0xdead0004,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000002,0xcafef00d,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000002,0x00000000,0xcafef00d,0xdead0002,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000002,0x00000000,0x00000000,0x00000000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0xdead0002,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000003,0xcafef00d,0xdead0002,0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000003,0x00000000,0xcafef00d,0x00000000,0xdead0002,0xcafef00d,0x00000000,0xdead0004,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000003,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000003,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000003,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0x00000003,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0002,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0002,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0002,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0002,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0xdead0004,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0x00000000,0xdead0004,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
   { 0xcafef00d,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0xdead0004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000003 },
};

int results2_pos = 0;

void dsp_test_end_results(u32* addr_ptr, int pos, int add_mode, int type, int hold, int length, int user_wait_display, int verify, u32 results[128][65])
{
   int i;
   SCU_REG_PDA = 0;

   u32 data[66] = { 0 };

   u32 dsp_data[64] = { 0 };

   for (i = 0; i < 64; i++)
      dsp_data[i] = SCU_REG_PDD;

   if (type == 1 || type == 3)
   {
      for (i = 0; i < 64; i++)
         data[i] = dsp_data[i];
   }
   else
   {
      for (i = 0; i < 64; i++)
         data[i] = addr_ptr[i];
   }

   SCU_REG_PDA = 2 << 6 | 0;

   u32 ct2 = SCU_REG_PDD;

   data[i++] = ct2;

   char result[2048] = { 0 };

   sprintf(result, "{");
   dsp_copy_str_to_offset(result);

   for (i = 0; i < 65; i++)
   {
      if (i == 64)
         sprintf(result, "0x%08x", data[i]);
      else
         sprintf(result, "0x%08x,", data[i]);

      dsp_copy_str_to_offset(result);
   }

   sprintf(result, "},\n");
   dsp_copy_str_to_offset(result);

   if (user_wait_display)
   {
      vdp_printf(&test_disp_font, 22 * 8, 0 * 8, 0xF, "DSP");
      vdp_printf(&test_disp_font, 31 * 8, 0 * 8, 0xF, "DSTA");
      vdp_printf(&test_disp_font, 13 * 8, 0 * 8, 0xF, "SRC/DST");

      for (i = 0; i < 27; i++)
      {
         vdp_printf(&test_disp_font, 22 * 8, (i + 1) * 8, 0xF, "%08X", dsp_data[i]);
         vdp_printf(&test_disp_font, 31 * 8, (i + 1) * 8, 0xF, "%08X", values[i]);

         if (i < 24)
         {
            vdp_printf(&test_disp_font, 13 * 8, (i + 1) * 8, 0xF, "%08X", addr_ptr[i + 24]);
            vdp_printf(&test_disp_font, 4 * 8, (i + 1) * 8, 0xF, "%08X", addr_ptr[i]);
         }
      }

      vdp_printf(&test_disp_font, 0 * 8, 25 * 8, 0xF, "%08X", ct2);

      vdp_printf(&test_disp_font, 0 * 8, 26 * 8, 0xF, "add %d type %d hold %d", add_mode, type, hold);

      for (;;)
      {
         vdp_vsync();
         ud_check(0);

         if (per[0].but_push_once & PAD_A)
         {
            break;
         }

         if (per[0].but_push_once & PAD_Y)
         {
            reset_system();
         }
      }
   }
   else
   {
      if (verify || AUTO)
      {
         int fail = 0;

         for (i = 0; i < 64; i++)
         {
            if (data[i] != results[results2_pos][i])
            {
               //if (!AUTO)
               {
                  test_disp_font.transparent = 0;
                  vdp_printf(&test_disp_font, 0 * 8, 26 * 8, 0xF, "%02X %08X should be %08X %08X", i, data[i], results[results2_pos][i], addr_ptr);
                  vdp_printf(&test_disp_font, 0 * 8, 27 * 8, 0xF, "add %d type %d hold %d result pos %d len %d", add_mode, type, hold, results2_pos, length);
               }

               fail = 1;

               int q;

               //if (!AUTO)
               {
                  for (q = 0; q < 26; q++)
                  {
                     vdp_printf(&test_disp_font, 20 * 8, q * 8, 0xF, "%02X %08X %08X", q, data[q], results[results2_pos][q]);
                  }
               }
               break;
            }
         }

         if (fail)
         {
            //if (!AUTO)
            {
               for (;;)
               {
                  vdp_vsync();

                  if (per[0].but_push_once & PAD_A)
                  {
                     break;
                  }

                  if (per[0].but_push_once & PAD_Y)
                  {
                     reset_system();
                  }
               }
            }

            stage_status = STAGESTAT_BADDATA;
         }

         results2_pos++;
      }
   }
}

void dsp_exec_read_dsta()
{
   int pos = 0;

   dsp_exec(0);

   while (dsp_is_exec())
   {
      u32 val = SCUREG_DSTA;

      if (val)
         values[pos++] = val;

      if (pos > 4094)
         break;
   }
}

void write_dma_instruction(int type, u32 bank, u32 add_mode, u8 length, u8 ct, int hold)
{
   u32 instruction = 0;
   if (type == 1)
      instruction = DMA_D0_RAM_Imm(add_mode, bank, length);
   else if (type == 2)
      instruction = DMA_RAM_D0_Imm(add_mode, bank, length);
   else if (type == 3)
      instruction = DMA_D0_RAM_s(add_mode, bank, ct);
   else if (type == 4)
      instruction = DMA_RAM_D0_s(add_mode, bank, ct);

   SCU_REG_PPD = instruction | (hold << 14);
}

void write_jump(int jump_loc)
{
   SCU_REG_PPD = JMP_T0_Imm(jump_loc);
   SCU_REG_PPD = NOP();
}

void write_wa0_ra0(int type, u32 addr)
{
   u32 ra0_wa0 = MVIDEST_RA0;

   if (type == 2 || type == 4)
      ra0_wa0 = MVIDEST_WA0;

   SCU_REG_PPD = MOV_Imm_d(0, MOVDEST_CT0);
   SCU_REG_PPD = MOV_Imm_d(0, MOVDEST_CT1);

   SCU_REG_PPD = MVI_Imm_d(addr >> 2, ra0_wa0);
}

void dsp_clear()
{
   int i, j;

   //clear all dma ram banks
   for (j = 0; j < 4; j++)
   {
      SCU_REG_PDA = j << 6 | 0;

      for (i = 0; i < 64; i++)
      {
         SCU_REG_PDD = 0;
      }
   }
}

void write_dma_to_dsp_data(u32* addr_ptr, u32 length)
{
   int i, j;

   dsp_clear();

   for (i = 0; i < length; i += 2)
   {
      addr_ptr[i] = 0xdead0000 | i;
      addr_ptr[i + 1] = 0xcafef00d;
   }
}

void write_dma_from_dsp_data(u32* addr_ptr, u32 length)
{
   int i;

   SCU_REG_PDA = 0;

   for (i = 0; i < length; i++)
   {
      addr_ptr[i] = 0;
   }

   for (i = 0; i < length; i += 2)
   {
      SCU_REG_PDD = 0xdead0000 | i;
      SCU_REG_PDD = 0xcafef00d;
   }
}

void write_test_data(u32* addr_ptr, int type, int addr_clear, u32 length)
{
   int i;

   if (addr_clear)
   {
      if (type == 1 || type == 3)
      {
         write_dma_to_dsp_data(addr_ptr, length);
      }
      else
      {
         write_dma_from_dsp_data(addr_ptr, length);
      }
   }
   else
   {
      for (i = 0; i < length; i += 2)
      {
         addr_ptr[i] = 0xdead0000 | i;
         addr_ptr[i + 1] = 0xcafef00d;
      }
   }
}

void dsp_init_stop(u32 length, int type)
{
   u32 testval = SCU_REG_PPAF;
   (void)testval;

   dsp_stop();

   SCU_REG_PPAF = 0x8000;
   SCU_REG_PDA = 0;

   if (type == 3 || type == 4)
   {
      SCU_REG_PDD = length;
   }
}

void dsp_dma_to_dsp(u32 src_dest_addr, u32 bank, u32 add_mode, u8 length, int type, int addr_clear, int hold, int user_wait_display, u32 results[128][65])
{
   u32* addr_ptr = (volatile u32*)(src_dest_addr);

   int i;

   write_test_data(addr_ptr, type, addr_clear, length);

   dsp_init_stop(length, type);

   write_wa0_ra0(type, src_dest_addr);

   write_dma_instruction(type, bank, add_mode, length, 0, hold);
   
   write_jump(3);

   SCU_REG_PPD = END();

   dsp_exec_read_dsta();

   dsp_test_end_results(addr_ptr, 0, add_mode, type, hold, length, user_wait_display, 0, results);
}

//back to back dma to test CT and RA0/WA0 regs
void dsp_dma_multiple(u32 addr, u32 bank, u32 add_mode, u8 length, int type, int ct, int hold, int user_wait_display, int verify, u32 results[128][65])
{
   u32* addr_ptr = (volatile u32*)(addr);

   write_test_data(addr_ptr, type, 1, 64);

   dsp_init_stop(length, type);

   if (type == 3 || type == 4)
   {
      int page = 1;
      int address = 0;
      SCU_REG_PDA = page << 6 | address;
      SCU_REG_PDD = length;
      SCU_REG_PDD = length + 1;
   }

   write_wa0_ra0(type, addr);

   write_dma_instruction(type, bank, add_mode, length, ct, hold);

   //save ct counter to see if it was incremented by ct++ dma
   SCU_REG_PPD = MOV_s_d(MOVSRC_MC1, MOVDEST_MC2);

   write_jump(5);

   write_dma_instruction(type, bank, add_mode, length, ct, hold);

   write_jump(8);

   SCU_REG_PPD = END();

   dsp_exec_read_dsta();

   dsp_test_end_results(addr_ptr, 0, add_mode, type, hold, length, user_wait_display, verify, results);
}

void test_dma_reads(u32 address, int user_wait_display, int verify, u32 results[128][65])
{
   int length = 2;
   int add_mode = 0;
   int count_bank = 1;
   int type = 1;
   int hold = 0;
   //type 1 / 5

   for (hold = 0; hold <= 1; hold++)
   {
      for (length = 2; length <= 3; length++)
      {
         for (add_mode = DMAADD_0; add_mode <= DMAADD_64; add_mode++)
         {
            dsp_dma_multiple(address, DMARAM_0, add_mode, length, 1, count_bank, hold, user_wait_display, verify, results);
         }
      }
   }

   //type 3 / 7
   for (hold = 0; hold <= 1; hold++)
   {
      for (length = 2; length <= 3; length++)
      {
         for (add_mode = DMAADD_0; add_mode <= DMAADD_64; add_mode++)
         {
            dsp_dma_multiple(address, DMARAM_0, add_mode, length, 3, count_bank, hold, user_wait_display, verify, results);
         }
      }
   }
}

void test_dma_writes(u32 address, int user_wait_display, int verify, u32 results[128][65])
{
   int length = 2;
   int add_mode = 0;
   int count_bank = 1;
   int type = 1;
   int hold = 0;

   //type 2 / 6
   for (hold = 0; hold <= 1; hold++)
   {
      for (length = 2; length <= 3; length++)
      {
         for (add_mode = DMAADD_0; add_mode <= DMAADD_64; add_mode++)
         {
            dsp_dma_multiple(address, DMARAM_0, add_mode, length, 2, count_bank, hold, user_wait_display, verify, results);
         }
      }
   }

   //type 4 / 8
   for (hold = 0; hold <= 1; hold++)
   {
      for (length = 2; length <= 3; length++)
      {
         for (add_mode = DMAADD_0; add_mode <= DMAADD_64; add_mode++)
         {
            dsp_dma_multiple(address, DMARAM_0, add_mode, length, 4, count_bank, hold, user_wait_display, verify, results);
         }
      }
   }
}

void cart_init()
{
   volatile u16 * init = (volatile u16 *)0x257EFFFE;
   init[0] = 1;
   SCU_REG_ABUSSRCS0CS1 = 0x23301FF0;
   SCU_REG_ABUSREFRESH = 0x00000013;

}

void dsp_dma_c_bus_read()
{
   results2_pos = 0;

   test_dma_reads(0x26070000, 0, 1, c_results);//high wram (c bus)

   if (stage_status != STAGESTAT_BADDATA)
      stage_status = STAGESTAT_DONE;
}

void dsp_dma_c_bus_write()
{
   test_dma_writes(0x26070000, 0, 1, c_results);

   if (stage_status != STAGESTAT_BADDATA)
      stage_status = STAGESTAT_DONE;
}

void dsp_dma_b_bus_read()
{
   results2_pos = 0;

   test_dma_reads(0x25c01000, 0, 1, b_results);//vdp1 ram (b bus)

   if (stage_status != STAGESTAT_BADDATA)
      stage_status = STAGESTAT_DONE;
}

void dsp_dma_b_bus_write()
{
   test_dma_writes(0x25c01000, 0, 1, b_results);

   if (stage_status != STAGESTAT_BADDATA)
      stage_status = STAGESTAT_DONE;
}

void dsp_dma_a_bus_read()
{
   cart_init();
   results2_pos = 0;


   test_dma_reads(0x22400000,0, 1, a_results);//cart (a bus)

   if (stage_status != STAGESTAT_BADDATA)
      stage_status = STAGESTAT_DONE;
}

void dsp_dma_a_bus_write()
{
   test_dma_writes(0x22400000,0, 1, a_results);

   if (stage_status != STAGESTAT_BADDATA)
      stage_status = STAGESTAT_DONE;
}

void dsp_print_banks(int start_x)
{
   int j, i;

   u32 banks[4][64] = { 0 };

   for (j = 0; j < 4; j++)
   {
      SCU_REG_PDA = j << 6 | 0;

      for (i = 0; i < 64; i++)
      {
         banks[j][i] = SCU_REG_PDD;
      }
   }

   for (i = 0; i < 24; i++)
   {
      vdp_printf(&test_disp_font, (start_x + 0) * 8, (i + 1) * 8, 0xF, "%08X", banks[0][i]);
      vdp_printf(&test_disp_font, (start_x + 9) * 8, (i + 1) * 8, 0xF, "%08X", banks[1][i]);
      vdp_printf(&test_disp_font, (start_x + 18) * 8, (i + 1) * 8, 0xF, "%08X", banks[2][i]);
      vdp_printf(&test_disp_font, (start_x + 27) * 8, (i + 1) * 8, 0xF, "%08X", banks[3][i]);
   }
}

void dsp_print_full_bank(int which, int start_x)
{
   int j, i;

   u32 bank[64] = { 0 };

   SCU_REG_PDA = which << 6 | 0;

   for (i = 0; i < 64; i++)
   {
      bank[i] = SCU_REG_PDD;
   }

   for (i = 0; i < 16; i++)
   {
      vdp_printf(&test_disp_font, (start_x + 0) * 8, (i + 1) * 8, 0xF, "%08X", bank[i+0]);
      vdp_printf(&test_disp_font, (start_x + 9) * 8, (i + 1) * 8, 0xF, "%08X", bank[i+16]);
      vdp_printf(&test_disp_font, (start_x + 18) * 8, (i + 1) * 8, 0xF, "%08X", bank[i+32]);
      vdp_printf(&test_disp_font, (start_x + 27) * 8, (i + 1) * 8, 0xF, "%08X", bank[i+48]);
   }
}

void dma_test_no_wait(int num_nops, int add_mode)
{
   u32 addr = 0x25c01000;
   int type = 1;
   int length = 64;
   int ct = 0;
   int hold = 0;

   u32* addr_ptr = (volatile u32*)(addr);

   write_test_data(addr_ptr, type, 1, 64);

   dsp_init_stop(length, type);

   write_wa0_ra0(type, addr);

   length = 8;

   SCU_REG_PPD = MOV_Imm_d(0, MOVDEST_CT2);
   SCU_REG_PPD = MOV_Imm_d(0, MOVDEST_CT3);

   write_dma_instruction(type, DMARAM_0, add_mode, length, ct, hold);

   int i;

   for (i = 0; i < num_nops; i++)
      SCU_REG_PPD = NOP();

   write_dma_instruction(type, DMARAM_1, add_mode, length, ct, hold);

   write_dma_instruction(type, DMARAM_2, add_mode, length, ct, hold);

   write_dma_instruction(type, DMARAM_3, add_mode, length, ct, hold);

   write_jump(9 + num_nops);

   SCU_REG_PPD = END();

   dsp_exec_read_dsta();

   if (!AUTO)
   {
      vdp_printf(&test_disp_font, 0 * 8, 27 * 8, 0xF, "%08X", num_nops);

      dsp_print_banks(0);

      for (;;)
      {
         vdp_vsync();
         ud_check(0);

         if (per[0].but_push_once & PAD_A)
         {
            break;
         }

         if (per[0].but_push_once & PAD_Y)
         {
            reset_system();
         }
      }
   }
}

//submit dmas without waiting for the first to complete
//demonstrates that multiple dmas can be queued with
//proper ra0 / wa0 increment
void dsp_multiple_submit()
{
   if(!AUTO)
      test_disp_font.transparent = 0;

   dma_test_no_wait(0, DMAADD_2);

   u32 bank[4] = { 0 };

   int j;

   for (j = 0; j < 4; j++)
   {
      SCU_REG_PDA = j << 6 | 0;

      bank[j] = SCU_REG_PDD;
   }

   if (AUTO)
   {
      if (bank[0] != 0xdead0000)
         stage_status = STAGESTAT_BADDATA;
      else if (bank[1] != 0xdead0008)
         stage_status = STAGESTAT_BADDATA;
      else if (bank[2] != 0xdead0010)
         stage_status = STAGESTAT_BADDATA;
      else if (bank[3] != 0xdead0018)
         stage_status = STAGESTAT_BADDATA;
      else
         stage_status = STAGESTAT_DONE;
   }
}

void dsp_dma_timing_impl(int length)
{
   u32 addr = 0x25c01000;
   int type = 1;
   int ct = 0;
   int hold = 0;
   int i;
   int add_mode = 1;

   u32* addr_ptr = (volatile u32*)(addr);

   write_test_data(addr_ptr, type, 1, 256);

   dsp_init_stop(length, type);

   write_wa0_ra0(type, addr);

   SCU_REG_PPD = MVI_Imm_d(0, MVIDEST_MC0);
   SCU_REG_PPD = MVI_Imm_d(1, MVIDEST_MC1);
   SCU_REG_PPD = MOV_Imm_d(0, MOVDEST_CT3);

   write_dma_instruction(type, DMARAM_3, add_mode, length, ct, hold);

   //reset counters
   SCU_REG_PPD = MOV_Imm_d(0, MOVDEST_CT0);
   SCU_REG_PPD = MOV_Imm_d(0, MOVDEST_CT1);

   //load acl and pl with md0 and md1 values
   SCU_REG_PPD = MOV_s_A(MOVSRC_MC0P) | MOV_s_P(MOVSRC_MC1P);

   SCU_REG_PPD = MOV_Imm_d(0, MOVDEST_CT0);
   SCU_REG_PPD = MOV_Imm_d(0, MOVDEST_CT1);

   //increment counter and move it to bank 0
   SCU_REG_PPD = ADD() | MOV_s_d(MOVSRC_ALUL, MOVDEST_MC0);

   //loop
   write_jump(7);

   SCU_REG_PPD = END();

   dsp_exec_read_dsta();

   if (!AUTO)
   {
      vdp_printf(&test_disp_font, 0 * 8, 27 * 8, 0xF, "%08X", length);

      dsp_print_banks(0);

      for (;;)
      {
         vdp_vsync();
         ud_check(0);

         if (per[0].but_push_once & PAD_A)
         {
            break;
         }

         if (per[0].but_push_once & PAD_Y)
         {
            reset_system();
         }
      }
   }
}

void dsp_dma_timing()
{
   int i;
   test_disp_font.transparent = 0;
   for (i = 0; i < 256; i += 16)
   {
      dsp_dma_timing_impl(i);
   }
}

//see if the dma count is added to ct immediately
void dma_ct_add(int add_mode, int type, u32 * bank)
{
   u32 addr = 0x25c01000;
   int length = 16;
   int ct = 0;
   int hold = 0;
   int i;

   u32* addr_ptr = (volatile u32*)(addr);

   write_test_data(addr_ptr, type, 1, 64);

   dsp_init_stop(length, type);

   write_wa0_ra0(type, addr);

   write_dma_instruction(type, DMARAM_0, add_mode, length, ct, hold);

   SCU_REG_PPD = MVI_Imm_d(0x00c0ffee, MVIDEST_MC0);

   write_jump(5);

   SCU_REG_PPD = END();

   dsp_exec_read_dsta();

   if(!AUTO)
      dsp_print_full_bank(0, 0);

   SCU_REG_PDA = 0 << 6 | 0;

   for (i = 0; i < 64; i++)
      bank[i] = SCU_REG_PDD;

   if (!AUTO)
   {
      for (;;)
      {
         vdp_vsync();
         ud_check(0);

         if (per[0].but_push_once & PAD_A)
         {
            break;
         }

         if (per[0].but_push_once & PAD_Y)
         {
            reset_system();
         }
      }
   }
}

void dma_ct_add_test1()
{
   if (!AUTO)
      test_disp_font.transparent = 0;

   u32 bank[64] = { 0 };

   dma_ct_add(1, 1, bank);

   if (AUTO)
   {
      if (bank[16] != 0x00c0ffee)
         stage_status = STAGESTAT_BADDATA;
      else
         stage_status = STAGESTAT_DONE;
   }
}

void dma_ct_add_test2()
{
   if (!AUTO)
      test_disp_font.transparent = 0;

   u32 bank[64] = { 0 };

   dma_ct_add(1, 2, bank);

   if (AUTO)
   {
      if (bank[16] != 0x00c0ffee)
         stage_status = STAGESTAT_BADDATA;
      else
         stage_status = STAGESTAT_DONE;
   }
}

int is_out_of_range(int target, int val)
{
   int fudge = 4;
   int max = target + fudge;
   int min = target - fudge;

   if (min <= 0)
      min = 1;

   if (val > max || val < min)
      return 1;

   return 0;
}

void check_dma_timing(int count, int expected)
{
   dsp_dma_timing_impl(count);

   u32 time = 0;
   SCU_REG_PDA = 0 << 6 | 0;
   time = SCU_REG_PDD;

   if (is_out_of_range(expected, time))
   {
      stage_status = STAGESTAT_BADDATA;
      vdp_printf(&test_disp_font, 0 * 8, 27 * 8, 0xF, "%08X", time);
   }
}

void dma_timing_256()
{
   check_dma_timing(0, 0x23);

   if (stage_status != STAGESTAT_BADDATA)
      stage_status = STAGESTAT_DONE;
}

void dma_timing_16()
{
   check_dma_timing(16, 5);

   if (stage_status != STAGESTAT_BADDATA)
      stage_status = STAGESTAT_DONE;
}

void dma_timing_32()
{
   check_dma_timing(32, 7);

   if (stage_status != STAGESTAT_BADDATA)
      stage_status = STAGESTAT_DONE;
}

void scu_dma_read_write_increment_a_to_b();
void scu_dma_read_write_increment_a_to_c();
void scu_dma_read_write_increment_b_to_a();
void scu_dma_read_write_increment_b_to_c();
void scu_dma_read_write_increment_c_to_a();
void scu_dma_read_write_increment_c_to_b();

void scu_dsp_dma_tests()
{
   if (AUTO)
   {
      unregister_all_tests();
      register_test(&dma_ct_add_test1, "CT reg += DMA count instantly (read)");
      register_test(&dma_ct_add_test2, "CT reg += DMA count instantly (write)");
      register_test(&dma_timing_256, "DMA timing 0 (256) words VDP1 RAM");
      register_test(&dma_timing_16, "DMA timing 16 words VDP1 RAM");
      register_test(&dma_timing_32, "DMA timing 32 words VDP1 RAM");
      register_test(&dsp_multiple_submit, "Multiple DMA submit RA0/WA0 check");
      register_test(&dsp_dma_c_bus_read, "DSP DMA C-bus Reads");
      register_test(&dsp_dma_c_bus_write, "DSP DMA C-bus Writes");
      register_test(&dsp_dma_b_bus_read, "DSP DMA B-bus Reads");
      register_test(&dsp_dma_b_bus_write, "DSP DMA B-bus Writes");
      register_test(&dsp_dma_a_bus_read, "DSP DMA A-bus Reads");
      register_test(&dsp_dma_a_bus_write, "DSP DMA A-bus Writes");
      register_test(&scu_dma_read_write_increment_a_to_b, "SCU DMA Read/Write Increment A->B");
      register_test(&scu_dma_read_write_increment_a_to_c, "SCU DMA Read/Write Increment A->C");
      register_test(&scu_dma_read_write_increment_b_to_a, "SCU DMA Read/Write Increment B->A");
      register_test(&scu_dma_read_write_increment_b_to_c, "SCU DMA Read/Write Increment B->C");
      register_test(&scu_dma_read_write_increment_c_to_a, "SCU DMA Read/Write Increment C->A");
      register_test(&scu_dma_read_write_increment_c_to_b, "SCU DMA Read/Write Increment C->B");
      do_tests("SCU DSP DMA tests", 0, 0);
   }
   else
   {

   }
}
u32 a_to_b_results[][8] =
{
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xbead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xbeef0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xbeef9000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xbead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xbeef0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xbeef9000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xbead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xbeef0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xbeef9000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xbead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xbeef0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xbead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xbeef0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xbeef9000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafe0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xf0fe0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xf0010000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xf0019000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xba5e0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xba5e0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xba020000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xba029000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xf01d0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xab1d0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xab030000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xde000000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdead0000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbe00,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0xde000000,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0xdead0000,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0xdeadbe00,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0xdeadbeef,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0xdeadbeef,0x00000002,0xde000000,0x00000003, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0xdeadbeef,0x00000002,0xdead0000,0x00000003, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0xdeadbeef,0x00000002,0xdeadbe00,0x00000003, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0xdeadbeef,0x00000002,0xdeadbeef,0x00000003, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xca000000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafe0000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba000000,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5e0000,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba00,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0xf0000000,0x00000003, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0xf01d0000,0x00000003, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0xf01dab00,0x00000003, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0xf01dab03,0x00000003, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbe000000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xbe009000,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xdead0000,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xdead0000,0x00000002,0xbe000000,0x00000003, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xdead0000,0x00000002,0xbeef0000,0x00000003, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xdead0000,0x00000002,0xbeef0000,0x00000003, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xdead0000,0x00000002,0xbeef0000,0x00000003, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xdead0000,0x00000002,0xbeef0000,0x00000003, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xdead0000,0x00000002,0xbeef0000,0x00000003, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xdead0000,0x00000002,0xbeef0000,0x00000003, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xdead0000,0x00000002,0xbeef0000,0x00000003, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xdead0000,0x00000002,0xbeef0000,0x00000003, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xdead0000,0x00000002,0xbeef0000,0x00000003, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbe000000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xbe009000,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xcafe0000,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xcafe0000,0x00000002,0xf0000000,0x00000003, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xcafe0000,0x00000002,0xf0010000,0x00000003, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xcafe0000,0x00000002,0xf0010000,0x00000003, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xcafe0000,0x00000002,0xf0010000,0x00000003, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xcafe0000,0x00000002,0xf0010000,0x00000003, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xcafe0000,0x00000002,0xf0010000,0x00000003, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xcafe0000,0x00000002,0xf0010000,0x00000003, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xcafe0000,0x00000002,0xf0010000,0x00000003, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xcafe0000,0x00000002,0xf0010000,0x00000003, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000001,0xcafe0000,0x00000002,0xf0010000,0x00000003, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbe000000,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xbe009000,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xdead0000,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xdead0000,0x00000004,0xbe000000,0x00000006, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xdead0000,0x00000004,0xbeef0000,0x00000006, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xdead0000,0x00000004,0xbeef0000,0x00000006, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xdead0000,0x00000004,0xbeef0000,0x00000006, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xdead0000,0x00000004,0xbeef0000,0x00000006, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xdead0000,0x00000004,0xbeef0000,0x00000006, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xdead0000,0x00000004,0xbeef0000,0x00000006, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xdead0000,0x00000004,0xbeef0000,0x00000006, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xdead0000,0x00000004,0xbeef0000,0x00000006, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xdead0000,0x00000004,0xbeef0000,0x00000006, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbe000000,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xbe009000,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xcafe0000,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xcafe0000,0x00000004,0xf0000000,0x00000006, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xcafe0000,0x00000004,0xf0010000,0x00000006, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xcafe0000,0x00000004,0xf0010000,0x00000006, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xcafe0000,0x00000004,0xf0010000,0x00000006, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xcafe0000,0x00000004,0xf0010000,0x00000006, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xcafe0000,0x00000004,0xf0010000,0x00000006, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xcafe0000,0x00000004,0xf0010000,0x00000006, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xcafe0000,0x00000004,0xf0010000,0x00000006, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xcafe0000,0x00000004,0xf0010000,0x00000006, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000002,0xcafe0000,0x00000004,0xf0010000,0x00000006, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbe000000,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xbe009000,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xdead0000,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xdead0000,0x00000008,0xbe000000,0x0000000c, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xdead0000,0x00000008,0xbeef0000,0x0000000c, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xdead0000,0x00000008,0xbeef0000,0x0000000c, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xdead0000,0x00000008,0xbeef0000,0x0000000c, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xdead0000,0x00000008,0xbeef0000,0x0000000c, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xdead0000,0x00000008,0xbeef0000,0x0000000c, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xdead0000,0x00000008,0xbeef0000,0x0000000c, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xdead0000,0x00000008,0xbeef0000,0x0000000c, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xdead0000,0x00000008,0xbeef0000,0x0000000c, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xdead0000,0x00000008,0xbeef0000,0x0000000c, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbe000000,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xbe009000,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xcafe0000,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xcafe0000,0x00000008,0xf0000000,0x0000000c, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xcafe0000,0x00000008,0xf0010000,0x0000000c, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xcafe0000,0x00000008,0xf0010000,0x0000000c, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xcafe0000,0x00000008,0xf0010000,0x0000000c, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xcafe0000,0x00000008,0xf0010000,0x0000000c, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xcafe0000,0x00000008,0xf0010000,0x0000000c, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xcafe0000,0x00000008,0xf0010000,0x0000000c, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xcafe0000,0x00000008,0xf0010000,0x0000000c, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xcafe0000,0x00000008,0xf0010000,0x0000000c, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000004,0xcafe0000,0x00000008,0xf0010000,0x0000000c, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbe000000,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xbe009000,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xdead0000,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xdead0000,0x00000010,0xbe000000,0x00000018, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xdead0000,0x00000010,0xbeef0000,0x00000018, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xdead0000,0x00000010,0xbeef0000,0x00000018, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xdead0000,0x00000010,0xbeef0000,0x00000018, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xdead0000,0x00000010,0xbeef0000,0x00000018, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xdead0000,0x00000010,0xbeef0000,0x00000018, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xdead0000,0x00000010,0xbeef0000,0x00000018, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xdead0000,0x00000010,0xbeef0000,0x00000018, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xdead0000,0x00000010,0xbeef0000,0x00000018, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xdead0000,0x00000010,0xbeef0000,0x00000018, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbe000000,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xbe009000,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xcafe0000,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xcafe0000,0x00000010,0xf0000000,0x00000018, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xcafe0000,0x00000010,0xf0010000,0x00000018, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xcafe0000,0x00000010,0xf0010000,0x00000018, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xcafe0000,0x00000010,0xf0010000,0x00000018, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xcafe0000,0x00000010,0xf0010000,0x00000018, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xcafe0000,0x00000010,0xf0010000,0x00000018, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xcafe0000,0x00000010,0xf0010000,0x00000018, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xcafe0000,0x00000010,0xf0010000,0x00000018, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xcafe0000,0x00000010,0xf0010000,0x00000018, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000008,0xcafe0000,0x00000010,0xf0010000,0x00000018, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbe000000,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xbe009000,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xdead0000,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xdead0000,0x00000020,0xbe000000,0x00000030, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xdead0000,0x00000020,0xbeef0000,0x00000030, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xdead0000,0x00000020,0xbeef0000,0x00000030, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xdead0000,0x00000020,0xbeef0000,0x00000030, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xdead0000,0x00000020,0xbeef0000,0x00000030, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xdead0000,0x00000020,0xbeef0000,0x00000030, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xdead0000,0x00000020,0xbeef0000,0x00000030, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xdead0000,0x00000020,0xbeef0000,0x00000030, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xdead0000,0x00000020,0xbeef0000,0x00000030, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xdead0000,0x00000020,0xbeef0000,0x00000030, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbe000000,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xbe009000,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xcafe0000,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xcafe0000,0x00000020,0xf0000000,0x00000030, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xcafe0000,0x00000020,0xf0010000,0x00000030, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xcafe0000,0x00000020,0xf0010000,0x00000030, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xcafe0000,0x00000020,0xf0010000,0x00000030, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xcafe0000,0x00000020,0xf0010000,0x00000030, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xcafe0000,0x00000020,0xf0010000,0x00000030, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xcafe0000,0x00000020,0xf0010000,0x00000030, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xcafe0000,0x00000020,0xf0010000,0x00000030, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xcafe0000,0x00000020,0xf0010000,0x00000030, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000010,0xcafe0000,0x00000020,0xf0010000,0x00000030, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbe000000,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xbe009000,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xdead0000,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xdead0000,0x00000040,0xbe000000,0x00000060, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xdead0000,0x00000040,0xbeef0000,0x00000060, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xdead0000,0x00000040,0xbeef0000,0x00000060, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xdead0000,0x00000040,0xbeef0000,0x00000060, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xdead0000,0x00000040,0xbeef0000,0x00000060, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xdead0000,0x00000040,0xbeef0000,0x00000060, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xdead0000,0x00000040,0xbeef0000,0x00000060, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xdead0000,0x00000040,0xbeef0000,0x00000060, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xdead0000,0x00000040,0xbeef0000,0x00000060, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xdead0000,0x00000040,0xbeef0000,0x00000060, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbe000000,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xbe009000,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xcafe0000,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xcafe0000,0x00000040,0xf0000000,0x00000060, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xcafe0000,0x00000040,0xf0010000,0x00000060, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xcafe0000,0x00000040,0xf0010000,0x00000060, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xcafe0000,0x00000040,0xf0010000,0x00000060, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xcafe0000,0x00000040,0xf0010000,0x00000060, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xcafe0000,0x00000040,0xf0010000,0x00000060, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xcafe0000,0x00000040,0xf0010000,0x00000060, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xcafe0000,0x00000040,0xf0010000,0x00000060, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xcafe0000,0x00000040,0xf0010000,0x00000060, },
   { 0xdead0000,0x00000000,0xbeef0000,0x00000020,0xcafe0000,0x00000040,0xf0010000,0x00000060, },
};

u32 a_to_c_results[][8] =
{
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcaadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafebeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef0ef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xbafef001,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xba5ef001,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xba5eba01,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xba5eba02,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xf05eba02,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xf01dba02,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xf01dab02,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xf01dab03,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xde000000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdead0000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbe00,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcaadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafebeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef0ef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xba000000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xba5e0000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xba5eba00,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xba5eba02,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xf05eba02,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xf01dba02,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xf01dab02,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xf01dab03,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xde000000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdead0000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbe00,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0xde000000,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0xdead0000,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0xdeadbe00,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0xdeadbeef,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0xdeadbeef,0x00000002,0xde000000,0x00000003, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0xdeadbeef,0x00000002,0xdead0000,0x00000003, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0xdeadbeef,0x00000002,0xdeadbe00,0x00000003, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000001,0xdeadbeef,0x00000002,0xdeadbeef,0x00000003, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xca000000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafe0000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba000000,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5e0000,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba00,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0xf0000000,0x00000003, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0xf01d0000,0x00000003, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0xf01dab00,0x00000003, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0xf01dab03,0x00000003, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xde000000,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdead0000,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbe00,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000002,0xde000000,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000002,0xdead0000,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000002,0xdeadbe00,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000002,0xdeadbeef,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000002,0xdeadbeef,0x00000004,0xde000000,0x00000006, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000002,0xdeadbeef,0x00000004,0xdead0000,0x00000006, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000002,0xdeadbeef,0x00000004,0xdeadbe00,0x00000006, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000002,0xdeadbeef,0x00000004,0xdeadbeef,0x00000006, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xca000000,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafe0000,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef000,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba000000,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5e0000,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5eba00,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5eba02,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5eba02,0x00000004,0xf0000000,0x00000006, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5eba02,0x00000004,0xf01d0000,0x00000006, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5eba02,0x00000004,0xf01dab00,0x00000006, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5eba02,0x00000004,0xf01dab03,0x00000006, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xde000000,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdead0000,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbe00,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000004,0xde000000,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000004,0xdead0000,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000004,0xdeadbe00,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000004,0xdeadbeef,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000004,0xdeadbeef,0x00000008,0xde000000,0x0000000c, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000004,0xdeadbeef,0x00000008,0xdead0000,0x0000000c, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000004,0xdeadbeef,0x00000008,0xdeadbe00,0x0000000c, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000004,0xdeadbeef,0x00000008,0xdeadbeef,0x0000000c, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xca000000,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafe0000,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef000,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba000000,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5e0000,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5eba00,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5eba02,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5eba02,0x00000008,0xf0000000,0x0000000c, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5eba02,0x00000008,0xf01d0000,0x0000000c, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5eba02,0x00000008,0xf01dab00,0x0000000c, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5eba02,0x00000008,0xf01dab03,0x0000000c, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xde000000,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdead0000,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbe00,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000008,0xde000000,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000008,0xdead0000,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000008,0xdeadbe00,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000008,0xdeadbeef,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000008,0xdeadbeef,0x00000010,0xde000000,0x00000018, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000008,0xdeadbeef,0x00000010,0xdead0000,0x00000018, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000008,0xdeadbeef,0x00000010,0xdeadbe00,0x00000018, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000008,0xdeadbeef,0x00000010,0xdeadbeef,0x00000018, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xca000000,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafe0000,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef000,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba000000,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5e0000,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5eba00,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5eba02,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5eba02,0x00000010,0xf0000000,0x00000018, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5eba02,0x00000010,0xf01d0000,0x00000018, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5eba02,0x00000010,0xf01dab00,0x00000018, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5eba02,0x00000010,0xf01dab03,0x00000018, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xde000000,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdead0000,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbe00,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000010,0xde000000,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000010,0xdead0000,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000010,0xdeadbe00,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000010,0xdeadbeef,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000010,0xdeadbeef,0x00000020,0xde000000,0x00000030, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000010,0xdeadbeef,0x00000020,0xdead0000,0x00000030, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000010,0xdeadbeef,0x00000020,0xdeadbe00,0x00000030, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000010,0xdeadbeef,0x00000020,0xdeadbeef,0x00000030, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xca000000,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafe0000,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef000,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba000000,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5e0000,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5eba00,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5eba02,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5eba02,0x00000020,0xf0000000,0x00000030, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5eba02,0x00000020,0xf01d0000,0x00000030, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5eba02,0x00000020,0xf01dab00,0x00000030, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5eba02,0x00000020,0xf01dab03,0x00000030, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xde000000,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdead0000,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbe00,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000020,0xde000000,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000020,0xdead0000,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000020,0xdeadbe00,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000020,0xdeadbeef,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000020,0xdeadbeef,0x00000040,0xde000000,0x00000060, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000020,0xdeadbeef,0x00000040,0xdead0000,0x00000060, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000020,0xdeadbeef,0x00000040,0xdeadbe00,0x00000060, },
   { 0xdeadbeef,0x00000000,0xdeadbeef,0x00000020,0xdeadbeef,0x00000040,0xdeadbeef,0x00000060, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xca000000,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafe0000,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef000,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba000000,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5e0000,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5eba00,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5eba02,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5eba02,0x00000040,0xf0000000,0x00000060, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5eba02,0x00000040,0xf01d0000,0x00000060, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5eba02,0x00000040,0xf01dab00,0x00000060, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5eba02,0x00000040,0xf01dab03,0x00000060, },
};

u32 b_to_a_results[][8] =
{
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcaadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafebeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef0ef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xbafef001,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xba5ef001,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xba5eba01,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xba5eba02,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xf05eba02,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xf01dba02,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xf01dab02,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xf01dab03,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcaadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafebeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef0ef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xbafef001,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xba5ef001,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xba5eba01,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xba5eba02,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xf05eba02,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xf01dba02,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xf01dab02,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xf01dab03,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcaadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafebeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef0ef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xba000000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xba5e0000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xba5eba00,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xba5eba02,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xf05eba02,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xf01dba02,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xf01dab02,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xf01dab03,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcaadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafebeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef0ef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xba000000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xba5e0000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xba5eba00,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xba5eba02,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xf05eba02,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xf01dba02,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xf01dab02,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xcafef001,0x00000000,0xf01dab03,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xca000000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafe0000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba000000,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5e0000,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba00,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0xf0000000,0x00000003, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0xf01d0000,0x00000003, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0xf01dab00,0x00000003, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0xf01dab03,0x00000003, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xca000000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafe0000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef000,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba000000,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5e0000,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba00,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0xf0000000,0x00000003, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0xf01d0000,0x00000003, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0xf01dab00,0x00000003, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000001,0xba5eba02,0x00000002,0xf01dab03,0x00000003, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xca000000,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafe0000,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef000,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba000000,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5e0000,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5eba00,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5eba02,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5eba02,0x00000004,0xf0000000,0x00000006, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5eba02,0x00000004,0xf01d0000,0x00000006, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5eba02,0x00000004,0xf01dab00,0x00000006, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5eba02,0x00000004,0xf01dab03,0x00000006, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xca000000,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafe0000,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef000,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba000000,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5e0000,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5eba00,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5eba02,0x00000004,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5eba02,0x00000004,0xf0000000,0x00000006, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5eba02,0x00000004,0xf01d0000,0x00000006, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5eba02,0x00000004,0xf01dab00,0x00000006, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000002,0xba5eba02,0x00000004,0xf01dab03,0x00000006, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xca000000,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafe0000,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef000,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba000000,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5e0000,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5eba00,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5eba02,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5eba02,0x00000008,0xf0000000,0x0000000c, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5eba02,0x00000008,0xf01d0000,0x0000000c, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5eba02,0x00000008,0xf01dab00,0x0000000c, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5eba02,0x00000008,0xf01dab03,0x0000000c, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xca000000,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafe0000,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef000,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba000000,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5e0000,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5eba00,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5eba02,0x00000008,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5eba02,0x00000008,0xf0000000,0x0000000c, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5eba02,0x00000008,0xf01d0000,0x0000000c, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5eba02,0x00000008,0xf01dab00,0x0000000c, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000004,0xba5eba02,0x00000008,0xf01dab03,0x0000000c, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xca000000,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafe0000,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef000,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba000000,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5e0000,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5eba00,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5eba02,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5eba02,0x00000010,0xf0000000,0x00000018, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5eba02,0x00000010,0xf01d0000,0x00000018, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5eba02,0x00000010,0xf01dab00,0x00000018, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5eba02,0x00000010,0xf01dab03,0x00000018, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xca000000,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafe0000,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef000,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba000000,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5e0000,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5eba00,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5eba02,0x00000010,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5eba02,0x00000010,0xf0000000,0x00000018, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5eba02,0x00000010,0xf01d0000,0x00000018, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5eba02,0x00000010,0xf01dab00,0x00000018, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000008,0xba5eba02,0x00000010,0xf01dab03,0x00000018, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xca000000,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafe0000,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef000,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba000000,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5e0000,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5eba00,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5eba02,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5eba02,0x00000020,0xf0000000,0x00000030, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5eba02,0x00000020,0xf01d0000,0x00000030, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5eba02,0x00000020,0xf01dab00,0x00000030, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5eba02,0x00000020,0xf01dab03,0x00000030, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xca000000,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafe0000,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef000,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba000000,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5e0000,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5eba00,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5eba02,0x00000020,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5eba02,0x00000020,0xf0000000,0x00000030, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5eba02,0x00000020,0xf01d0000,0x00000030, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5eba02,0x00000020,0xf01dab00,0x00000030, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000010,0xba5eba02,0x00000020,0xf01dab03,0x00000030, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xca000000,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafe0000,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef000,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba000000,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5e0000,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5eba00,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5eba02,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5eba02,0x00000040,0xf0000000,0x00000060, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5eba02,0x00000040,0xf01d0000,0x00000060, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5eba02,0x00000040,0xf01dab00,0x00000060, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5eba02,0x00000040,0xf01dab03,0x00000060, },
   { 0xde000000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdead0000,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbe00,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xca000000,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafe0000,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef000,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0x0bad0bad,0x0bad0bad,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba000000,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5e0000,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5eba00,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5eba02,0x00000040,0x0bad0bad,0x0bad0bad, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5eba02,0x00000040,0xf0000000,0x00000060, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5eba02,0x00000040,0xf01d0000,0x00000060, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5eba02,0x00000040,0xf01dab00,0x00000060, },
   { 0xdeadbeef,0x00000000,0xcafef001,0x00000020,0xba5eba02,0x00000040,0xf01dab03,0x00000060, },
};

//a_to_b == c_to_b

//a_to_c == c_to_a

//b_to_a == b_to_c


#define MODE_MAKE_DATA 1
#define MODE_VERIFY 2
#define INC_AUTO 1


void scu_dma_read_write_increment_impl(
   u32 src_addr, u32 dest_addr, u32 length, u32 read_add, 
   u32 write_add, u32 write_address_update, u32 read_address_update, int mode, int results_pos, u32 my_results[][8])
{
   volatile u32 *src_addr_ptr = (volatile u32*)src_addr;
   volatile u32 *dest_addr_ptr = (volatile u32*)dest_addr;
   int i; 

   for (i = 0; i < 4096; i++)
   {
      src_addr_ptr[i] = 0;
   }

   for (i = 0; i < 256; i += 4)
   {
      if(i == 0)
         src_addr_ptr[i] = 0xdeadbeef;
      else
         src_addr_ptr[i] = 0xdeadbe00 | i;

      src_addr_ptr[i + 1] = 0xcafef000 | (i + 1);
      src_addr_ptr[i + 2] = 0xba5eba00 | (i + 2);
      src_addr_ptr[i + 3] = 0xf01dab00 | (i + 3);
   }

   for (i = 0; i < 4096; i++)
   {
      dest_addr_ptr[i] = 0;
   }

   SCUREG_D0R = src_addr;
   SCUREG_D0W = dest_addr;
   SCUREG_D0C = length;
   SCUREG_D0AD = (read_add << 8) | write_add;
   SCUREG_D0MD = 0x00000007 | (write_address_update << 8) | (read_address_update << 16);
   SCUREG_D0EN = 0x101;

   while (SCUREG_DSTA != 0) {}

   vdp_vsync();

   if (!INC_AUTO)
   {

      for (i = 0; i < 16; i++)
      {
         vdp_printf(&test_disp_font, 0 * 8, i * 8, 0xF, "%08X", src_addr_ptr[i]);
         vdp_printf(&test_disp_font, 9 * 8, i * 8, 0xF, "%08X", dest_addr_ptr[i]);
         vdp_printf(&test_disp_font, 18 * 8, i * 8, 0xF, "%08X", dest_addr_ptr[i + 16]);
         vdp_printf(&test_disp_font, 27 * 8, i * 8, 0xF, "%08X", dest_addr_ptr[i + 32]);
         vdp_printf(&test_disp_font, 36 * 8, i * 8, 0xF, "%08X", dest_addr_ptr[i + 48]);
      }

      vdp_printf(&test_disp_font, 0 * 8, 25 * 8, 0xF, "length   %08X ", length);
      vdp_printf(&test_disp_font, 0 * 8, 26 * 8, 0xF, "src_addr %08X dest_addr %08X", src_addr, dest_addr);
      vdp_printf(&test_disp_font, 0 * 8, 27 * 8, 0xF, "read_add %08X write_add %08X", read_add, write_add);
   }
#define LEN 255

   u32 data[LEN] = { 0 };

   for (i = 0; i < LEN; i++)
      data[i] = dest_addr_ptr[i];

   //store in a compressed format since the data is huge
   if (mode == MODE_MAKE_DATA)
   {
      char result_str[2048] = { 0 };
      sprintf(result_str, "{");
      dsp_copy_str_to_offset(result_str);

      int num_printed = 0;
      int num_to_print = 4;

      for (i = 0; i < LEN; i++)
      {
         if ((num_printed < num_to_print) && data[i] != 0)
         {
            sprintf(result_str, "0x%08x,", data[i]);
            dsp_copy_str_to_offset(result_str);
            sprintf(result_str, "0x%08x,", i);
            dsp_copy_str_to_offset(result_str);
            num_printed++;
         }
      }

      //fill unused space
      while (num_printed < num_to_print)
      {
         sprintf(result_str, "0x%08x,", 0x0bad0bad);
         dsp_copy_str_to_offset(result_str);
         sprintf(result_str, "0x%08x,", 0x0bad0bad);
         dsp_copy_str_to_offset(result_str);
         num_printed++;
      }

      sprintf(result_str, "},\n");
      dsp_copy_str_to_offset(result_str);
   }

   int fail = 0;

   if (mode == MODE_VERIFY)
   {
      for (i = 0; i < 8; i += 2)
      {
         //check nonzero values

         u32 val = my_results[results_pos][i];
         u32 pos = my_results[results_pos][i + 1];

         if (val == 0x0bad0bad && pos == 0x0bad0bad)
            break;

         if (data[pos] != val)
         {
            fail = 1;
            vdp_printf(&test_disp_font, 0 * 8, 24 * 8, 0xF, "%02X %08X should be %08X", i, data[pos], val);
         }

         //check gaps
         if (i < 6)
         {
            u32 pos1 = my_results[results_pos][i + 1];
            u32 pos2 = my_results[results_pos][i + 3];

            if (pos1 != 0xbad0bad && pos2 != 0x0bad0bad)
            {
               int j;

               for (j = pos1 + 1; j < pos2; j++)
               {
                  if (data[j] != 0)
                  {
                     fail = 1;
                     vdp_printf(&test_disp_font, 0 * 8, 24 * 8, 0xF, "%02X %08X should be %08X (GAP)", i, data[pos], 0);
                  }
               }
            }
         }
      }
   }

   if (fail)
      stage_status = STAGESTAT_BADDATA;

   if (!INC_AUTO)
   {
      if (fail || mode == 0)
      {
         for (;;)
         {
            vdp_vsync();
            ud_check(0);

            if (per[0].but_push_once & PAD_A)
            {
               break;
            }

            if (per[0].but_push_once & PAD_Y)
            {
               reset_system();
            }
         }
      }
   }
}

static const u32 a_bus_addr = 0x22400000;//cart
static const u32 b_bus_addr = 0x25C00000 + 0x1000;//vdp1 ram
static const u32 c_bus_addr = 0x260F0000 + 0x1000;//high wram

const int max_length = 16;
int mode = MODE_VERIFY;
//int mode = MODE_MAKE_DATA;

void scu_dma_read_write_increment_a_to_b()
{
   int write_add, read_add, length;
   int results_pos = 0;
   dsp_copy_str_to_offset("\r\nu32 a_to_b_results[][8] = \r\n{\r\n");

   for (write_add = 0; write_add <= 7; write_add++)
   {
      for (read_add = 0; read_add <= 1; read_add++)
      {
         for (length = 1; length <= max_length; length++)
         {
            scu_dma_read_write_increment_impl(a_bus_addr, b_bus_addr, length, read_add, write_add, 0, 0, mode, results_pos++, a_to_b_results);
         }
      }
   }

   dsp_copy_str_to_offset("};\r\n");

   if (stage_status != STAGESTAT_BADDATA)
      stage_status = STAGESTAT_DONE;
}

void scu_dma_read_write_increment_a_to_c()
{
   int write_add, read_add, length;
   int results_pos = 0;
   dsp_copy_str_to_offset("\r\nu32 a_to_c_results[][8] = \r\n{\r\n");

   for (write_add = 0; write_add <= 7; write_add++)
   {
      for (read_add = 0; read_add <= 1; read_add++)
      {
         for (length = 1; length <= max_length; length++)
         {
            scu_dma_read_write_increment_impl(a_bus_addr, c_bus_addr, length, read_add, write_add, 0, 0, mode, results_pos++, a_to_c_results);
         }
      }
   }

   dsp_copy_str_to_offset("};\r\n");

   if (stage_status != STAGESTAT_BADDATA)
      stage_status = STAGESTAT_DONE;
}

void scu_dma_read_write_increment_b_to_a()
{
   int write_add, read_add, length;
   int results_pos = 0;
   dsp_copy_str_to_offset("\r\nu32 b_to_a_results[][8] = \r\n{\r\n");

   for (write_add = 0; write_add <= 7; write_add++)
   {
      for (read_add = 0; read_add <= 1; read_add++)
      {
         for (length = 1; length <= max_length; length++)
         {
            scu_dma_read_write_increment_impl(b_bus_addr, a_bus_addr, length, read_add, write_add, 0, 0, mode, results_pos++, b_to_a_results);
         }
      }
   }

   dsp_copy_str_to_offset("};\r\n");

   if (stage_status != STAGESTAT_BADDATA)
      stage_status = STAGESTAT_DONE;
}

void scu_dma_read_write_increment_b_to_c()
{
   int write_add, read_add, length;
   int results_pos = 0;
   dsp_copy_str_to_offset("\r\nu32 b_to_c_results[][8] = \r\n{\r\n");

   for (write_add = 0; write_add <= 7; write_add++)
   {
      for (read_add = 0; read_add <= 1; read_add++)
      {
         for (length = 1; length <= max_length; length++)
         {
            scu_dma_read_write_increment_impl(b_bus_addr, c_bus_addr, length, read_add, write_add, 0, 0, mode, results_pos++, b_to_a_results);
         }
      }
   }

   dsp_copy_str_to_offset("};\r\n");

   if (stage_status != STAGESTAT_BADDATA)
      stage_status = STAGESTAT_DONE;
}

void scu_dma_read_write_increment_c_to_a()
{
   int write_add, read_add, length;
   int results_pos = 0;
   dsp_copy_str_to_offset("\r\nu32 c_to_a_results[][8] = \r\n{\r\n");

   for (write_add = 0; write_add <= 7; write_add++)
   {
      for (read_add = 0; read_add <= 1; read_add++)
      {
         for (length = 1; length <= max_length; length++)
         {
            scu_dma_read_write_increment_impl(c_bus_addr, a_bus_addr, length, read_add, write_add, 0, 0, mode, results_pos++, a_to_c_results);
         }
      }
   }

   dsp_copy_str_to_offset("};\r\n");

   if (stage_status != STAGESTAT_BADDATA)
      stage_status = STAGESTAT_DONE;
}

void scu_dma_read_write_increment_c_to_b()
{
   int write_add, read_add, length;
   int results_pos = 0;
   dsp_copy_str_to_offset("\r\nu32 c_to_b_results[][8] = \r\n{\r\n");

   for (write_add = 0; write_add <= 7; write_add++)
   {
      for (read_add = 0; read_add <= 1; read_add++)
      {
         for (length = 1; length <= max_length; length++)
         {
            scu_dma_read_write_increment_impl(c_bus_addr, b_bus_addr, length, read_add, write_add, 0, 0, mode, results_pos++, a_to_b_results);
         }
      }
   }

   dsp_copy_str_to_offset("};\r\n");

   if (stage_status != STAGESTAT_BADDATA)
      stage_status = STAGESTAT_DONE;
}



void scu_dma_read_write_increment_test()
{
   test_disp_font.transparent = 0;

   int write_add = 0;
   int read_add = 0;

   cart_init();

   if (mode == MODE_MAKE_DATA)
   {
      for (;;)
      {
         vdp_vsync();
         ud_check(0);

         if (per[0].but_push_once & PAD_A)
         {
            break;
         }

         if (per[0].but_push_once & PAD_Y)
         {
            reset_system();
         }
      }
   }

   cart_init();
   int length = 1;

   //a -> b
   scu_dma_read_write_increment_a_to_b();

   //a -> c
   scu_dma_read_write_increment_a_to_c();

   //b -> a
   scu_dma_read_write_increment_b_to_a();

   //b -> c (equal to b -> a)
   scu_dma_read_write_increment_b_to_c();

   //c -> a (equal to a -> c)
   scu_dma_read_write_increment_c_to_a();

   //c -> b (equal to a -> b)
   scu_dma_read_write_increment_c_to_b();


   if (stage_status != STAGESTAT_BADDATA)
      stage_status == STAGESTAT_DONE;

   if (!INC_AUTO)
   {
      for (;;)
      {
         vdp_vsync();
         ud_check(0);

         if (per[0].but_push_once & PAD_A)
         {
            break;
         }

         if (per[0].but_push_once & PAD_Y)
         {
            reset_system();
         }
      }
   }
}